/*
Copyright (C)2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define CONVERT_C
#include "bf_agent.h"

size_t convert_utf8_find_boundary(const char *in, size_t inlen)
{
	register uint8_t c;
	size_t len = inlen;
#ifdef ZOS
	char p[3];
#endif

	if (inlen <= 0)
		return 0;

	if (inlen > 3) {
		in = &in[inlen - 3];
		len = 3;
	}

#ifdef ZOS
	memcpy(p, in, len);
	__etoa_l(p, len);
	in = p;
#endif

	/* We can process the entire buffer in all cases except where
	 * the buffer ends in an incomplete (but so far valid) multi-byte
	 * sequence.  That is, any of these cases:
	 *   ...     ...     ...    C0-F4
	 *   ...     ...    E0-F4   80-BF
	 *   ...    F0-F4   80-BF   80-BF
	 * For each of these cases, the character boundary is immediately
	 * before the first character of the match.
	 */

	/* Check the last byte ... */
	if ((c = (uint8_t)(in[len - 1])) >= 0x80) {
		if (c >= 0xC0) return (c >= 0xF5) ? inlen : (inlen - 1);

		/* It was an extension byte, so check the next-to-last byte ... */
		if (len >= 2 && (c = (uint8_t)(in[len - 2])) >= 0x80) {
			if (c >= 0xE0) return (c >= 0xF5) ? inlen : (inlen - 2);
			if (c >= 0xC0) return inlen;

			/* It was an extension byte, so yet another check ... */
			if (len >= 3 && (c = (uint8_t)(in[len - 3])) >= 0x80) {
				if (c >= 0xF0) return (c >= 0xF5) ? inlen : (inlen - 3);
				if (c >= 0xC0) return inlen;
			}
		}
	}

	return inlen;
}

#define UTF8_REPLACE (0x3F)
#define UTF8_IS_EXT(c) (((c) & 0xC0) == 0x80)
unsigned int convert_utf8_clean(char *buf, size_t len, int replace)
{
	unsigned int errors = 0;
	uint8_t *p = (uint8_t *)buf;
	uint8_t *stop = (uint8_t *)&buf[len];

#ifdef ZOS
	__etoa_l(buf, len);
#endif

	while (p < stop) {
		/* ASCII */
		if (*p < 0x80) {
			++p;
			continue;
		}

		/* This catches:
		 * - extension byte that isn't part of a well-formed sequence
		 * - overlong encoding of ASCII as 2-byte
		 * - >4 byte sequence, which has to be overlong or exceed U+10FFFF
		 */
		if (*p < 0xC2 || *p >= 0xF5)
			goto malformed;

		/* 2-byte sequence */
		if (*p < 0xE0) {
			/* incomplete 2-byte sequence */
			if (p + 2 > stop || !UTF8_IS_EXT(p[1]))
				goto malformed;

			/* well-formed 2-byte sequence */
			p += 2;
			continue;
		}

		/* 3-byte sequence */
		if (*p < 0xF0) {
			/* This catches:
			 * - incomplete 3-byte sequence
			 * - overlong encoding of ASCII or 2-byte code point
			 * - code point is a UTF-16 surrogate (U+D800 - U+DFFF)
			 * - code point is U+FFFE or U+FFFF
			 */
			if (p + 3 > stop || !UTF8_IS_EXT(p[1]) || !UTF8_IS_EXT(p[2])
					|| (*p == 0xE0 && p[1] < 0xA0)
					|| (*p == 0xED && p[1] >= 0xA0)
					|| (*p == 0xEF && p[1] == 0xBF && p[2] >= 0xBE))
				goto malformed;

			/* well-formed 3-byte sequence */
			p += 3;
			continue;
		}

		/* 4-byte sequence. */
		/* This catches:
		 * - incomplete 4-byte sequence
		 * - overlong encoding of ASCII, 2-byte, or 3-byte code point
		 * - code point exceeds U+10FFFF
		 * - code point is U+xxFFFE or U+xxFFFF
		 */
		if (p + 4 > stop || !UTF8_IS_EXT(p[1])
				|| !UTF8_IS_EXT(p[2]) || !UTF8_IS_EXT(p[3]) 
				|| (*p == 0xF0 && p[1] < 0x90)
				|| (*p == 0xF4 && p[1] >= 0x90)
				|| ((p[1] & 0xF) == 0xF && p[2] == 0xBF && p[3] >= 0xBE))
			goto malformed;

		/* well-formed 4-byte sequence */
		p += 4;
		continue;

malformed:
		++errors;
		if (replace)
			*p++ = UTF8_REPLACE;
	}
#ifdef ZOS
	__atoe_l(buf, len);
#endif
	return errors;
}


/* For native encodings, we should be able to assume that we are at
 * a character boundary when we see a value less than 0x20, because
 * all of the character sets leave the control character range intact.
 * We don't know anything else about the structure of the code page,
 * so we make no other assumptions.
 */
size_t convert_native_find_boundary(const char *in, size_t inlen)
{
	register const char *p;

	for (p = &in[inlen-1]; p >= in; --p) {
		if ((unsigned char)(*p) < 0x20)
			return (p - in) + 1;
	}
	return 0;
}

#ifdef WIN32

/* We can assume a wide character boundary anywhere except after a high
 * surrogate for which we do not know the next value.
 */
size_t convert_wide_find_boundary(const wchar_t *in, size_t inlen)
{
	if (inlen > 0) {
		wchar_t c = in[inlen - 1];
		return (c < 0xD800 || c >= 0xDC00) ? inlen : (inlen - 1);
	}
	else return 0;
}
#endif /* WIN32 */

