/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/
#ifndef WIN32

#include "bf_agent.h"

int bfagent(void);

int daemon_done = 0;

#define MAX_LISTENERS 2
typedef struct {
	char *addrs[MAX_LISTENERS];
	struct pollfd fds[MAX_LISTENERS];
	int nfds;
} listener_queue_t;

static void bfperror(const char *s) {
	printf("[%d] bfdaemon: %s: %s\n",
			(int)getpid(), s, strerror(errno) );
}
static void bfsockerror(const char *addr, const char *port, const char *s) {
	printf("[%d] bfdaemon: [%s/%s]: %s: %s\n",
			(int)getpid(), addr, port, s, strerror(errno) );
}

static void close_listeners(listener_queue_t *listeners) {
	int i;
	for (i=0; i<listeners->nfds; ++i) {
		close(listeners->fds[i].fd);
		free(listeners->addrs[i]);
	}
	memset(listeners, 0, sizeof(*listeners));
}

static pid_t background(void) {
	pid_t pid = fork();
	switch (pid) {
		case -1:
			bfperror("fork");
			break;
		case 0:
			close(0);
			close(1);
			close(2);
			setsid();
	}
	return pid;
}


/* i5/OS seems to inexplicably throw EAPAR, and refuse to translate it,
 * if IPv6 is not configured.  This gives the annoying message:
 *   [pid] bfdaemon: [::/5555]: bind: Error 3418 occurred.
 * To silence it, we treat EAPAR the same way we treat EADDRINUSE;
 * that is, we ignore it if we already have at least one listener.
 */
#ifdef ISERIES_AGENT
#  ifndef EAPAR
#    define EAPAR 3418
#  endif /* !def EAPAR */
#  define IsPortUnavailable(x) (((x) == EADDRINUSE) || (x) == EAPAR)
# else /* !ISERIES_AGENT */
#  define IsPortUnavailable(x) ((x) == EADDRINUSE)
#endif /* ISERIES_AGENT/! */

static void create_listeners(const char *port, listener_queue_t *listeners)
{
	struct addrinfo hints;
	struct addrinfo *ai_head = NULL;
	struct addrinfo *ai;
	int one = 1;
	int ret;

	memset(listeners, 0, sizeof(*listeners));
	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
	hints.ai_socktype = SOCK_STREAM;
	{
		char *bind_addr = bfconf_get("bind", NULL);
		ret = getaddrinfo(bind_addr, port, &hints, &ai_head);
		if (NULL != bind_addr) { free(bind_addr); }
	}
	if (0 != ret) {
		printf("[%d] bfdaemon: %d: %s\n",
				(int)getpid(), ret, gai_strerror(ret) );
		return;
	}

	/* Go through all available interfaces where we could conceivably
	 * listen for stream connections.  This will presumably consist of
	 * IPv6 :: 5555 and IPv4 0.0.0.0 5555, in that order, subject
	 * to availability.
	 *
	 * The results from getaddrinfo are, according to RFC 3484, supposed
	 * to be sorted in an order of preference.  The system administrator
	 * is also supposed to be able to tune these preferences to their
	 * liking, such that they can prefer IPv4 to IPv6 is they so choose.
	 * On Linux, this is handled by /etc/gai.conf.
	 *
	 * This order is particularly important because some implementations
	 * (again Linux is a notable example) do not allow you to listen on
	 * both 0.0.0.0 (INADDR_ANY) and :: (IPv6's equivalent).  Whichever
	 * you ask for first is the winner, and the other bind request will
	 * fail with EADDRINUSE.
	 */
	for (ai = ai_head;
			NULL != ai && listeners->nfds < MAX_LISTENERS;
			ai = ai->ai_next)
	{
		int idx;
		char addr[128];
		int sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);

		if (sock == -1) {
			bfperror("socket");
			continue;
		}

		if (0 != getnameinfo(ai->ai_addr, ai->ai_addrlen,
					addr, sizeof(addr), NULL, 0, NI_NUMERICHOST))
		{
			sprintf(addr, "???[errno=%d]:%s", errno, port);
		}

		/* Set SO_REUSEADDR */
		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

		/* Attempt to bind to the port.  As described above, this can fail
		 * with EADDRINUSE if we already called bind for that port but a
		 * different address family.  We only report the error if we are
		 * debugging or if we have yet to succeed in creating a listener.
		 */
		if (0 != bind(sock, ai->ai_addr, ai->ai_addrlen)) {
			if (bf_debug || IsPortUnavailable(errno) || listeners->nfds == 0)
			{
				bfsockerror(addr, port, "bind");
			}
			close(sock);
			continue;
		}

		/* Create the listening queue */
		if (0 != listen(sock, SOMAXCONN)) {
			bfsockerror(addr, port, "listen");
			close(sock);
			continue;
		}

		/* Add this socket to the poll list */
		idx = listeners->nfds;
		listeners->addrs[idx] = strdup(addr);
		listeners->fds[idx].fd = sock;
		listeners->fds[idx].events = POLLIN;
		listeners->nfds++;

		/* Report the listener */
		if (bf_debug) {
			printf("[%d] bfdaemon: [%s/%s]: listening for connections\n",
					(int)getpid(), addr, port );
		}
	}
	freeaddrinfo(ai_head);
}

/* Run the agent with I/O redirected to the specified fd. */
static void start_agent(int fd) {
	int rc;

	/* Release SIGCHLD trap; otherwise, grantpt(2) could fail. */
	platform_signals_release();

	/* Redirect I/O */
	if (fd != 0) dup2(fd, 0);
	if (fd != 1) dup2(fd, 1);
	if (fd != 2) dup2(fd, 2);
	if (fd > 2) close(fd);

	/* Act like a normal agent process... */
	setsid();
	bfconf_refresh();
	rc = bfagent();
	shutdown(0, SHUT_RDWR);
	exit(rc);
}

/* Poll for incoming connections, accepting them as we see them... */
static void accept_connections(listener_queue_t *listeners, const char *port)
{
	int i, n, fd;
	char peeraddr[128];
	char peerport[32];
	struct pollfd *listener;
	struct linger lopts;
	struct sockaddr_storage peer;
	socklen_t peerlen;
	pid_t pid;

	n = poll(listeners->fds, listeners->nfds, -1);
	if (n <= 0) { return; }

	for (i=0; i<n; ++i) {
		listener = &listeners->fds[i];
		if (0 == (listener->revents & POLLIN)) { continue; }

		peerlen = sizeof(peer);
		fd = accept(listener->fd, (struct sockaddr *)&peer, &peerlen);
		if (fd == -1) {
			if (bf_debug) { bfperror("accept"); }
			continue;
		}

		if (0 != getnameinfo(
				(struct sockaddr *) &peer, peerlen,
				peeraddr, sizeof(peeraddr),
				peerport, sizeof(peerport),
				NI_NUMERICHOST | NI_NUMERICSERV ))
		{
			int err = errno;
			sprintf(peeraddr, "???:errno=%d", err);
			sprintf(peerport, "???:errno=%d", err);
		}

		if (!checkIP(peeraddr)) {
			close(fd);
			if (bf_debug) {
				printf("[%d] bfdaemon: [%s/%s]: rejected (see allow list)\n",
						(int)getpid(), peeraddr, peerport);
			}
			continue;
		}

		if (bf_debug) {
			char peername[NI_MAXHOST];
			if (0 != getnameinfo((struct sockaddr *) &peer, peerlen,
					peername, sizeof(peername), NULL, 0, 0))
			{
				sprintf(peername, "???:errno=%d", errno);
			}
			printf("[%d] bfdaemon: [%s/%s]: "
					"accepted connection from [%s/%s] (%s)\n",
					(int)getpid(), listeners->addrs[i], port,
					peeraddr, peerport, peername);
		}

		lopts.l_onoff = 1;
		lopts.l_linger = 20;
		setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *)&lopts, sizeof(lopts));

		pid = fork();
		switch (pid) {
			case -1:
				if (bf_debug)
					bfperror("fork");
				close(fd);
				continue;

			case 0:
				/* Child process; run the agent with redirected I/O ... */
				if (bf_debug) {
					printf("[%d] bfagent: launched\n", (int)getpid());
				}
				close_listeners(listeners);
				start_agent(fd);
				continue;
		}

		/* Parent process; continue polling ... */
		if (bf_debug) {
			printf("[%d] bfdaemon: [%s/%s]: attached to bfagent pid=%d\n",
					(int)getpid(), peeraddr, peerport, (int)pid );
		}
		close(fd);
	}
}



void bfdaemon(void) {
	listener_queue_t listeners;
	char *port;

	if (NULL == (port = bfconf_get("port", NULL)) ) {
		port = strdup("5555");
	}
	create_listeners(port, &listeners);
	if (listeners.nfds == 0) {
		printf("[%d] bfdaemon: no listeners; exiting\n", (int)getpid() );
		return;
	}

	/* If we aren't in debug mode, then it's time to go into the
	 * background...
	 */
	if (!bf_debug) {
		if (0 != background())
			return;
	}

	platform_signals_catch("bfdaemon");

	daemon_done = 0;
    while (!daemon_done) {
		accept_connections(&listeners, port);
	}

	close_listeners(&listeners);
	if (bf_debug) {
		printf("[%d] bfdaemon: exiting\n", (int)getpid());
	}
}


#endif /* !WIN32 */

