/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

extern const char *AGENT_VERSION;

static char *valid_var_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789-?";

static hash_ctx *env_hash_ctx = NULL;
static hash_table *env_ht_encrypted = NULL;
static hash_table *env_ht_hidden = NULL;
static hash_table *env_ht_recur = NULL;

static char *copy_var_name(const char *name) {
	char *p = strchr(name, '=');
	size_t idx = (NULL != p) ? (p - name) : strlen(name);

	p = malloc(idx + 1);
	memcpy(p, name, idx);
	p[idx] = '\0';
	return p;
}

void env_clear(void) {
	if (NULL != env_ht_hidden) {
		hash_destroy(env_ht_hidden);
		env_ht_hidden = NULL;
	}
	if (NULL != env_ht_encrypted) {
		hash_destroy(env_ht_encrypted);
		env_ht_encrypted = NULL;
	}
}

int env_is_hidden(const char *name) {
	if (NULL != env_ht_hidden) {
		char *p = copy_var_name(name);
		int ret = hash_contains_key_str(env_ht_hidden, p);
		free(p);
		return ret;
	}
	return 0;
}

void env_set_hidden(const char *name) {
	char *p;

	if (NULL == env_ht_hidden) {
		if (NULL == env_hash_ctx)
			env_hash_ctx = hash_ctx_new();
		env_ht_hidden = hash_new(env_hash_ctx, 11); 
	}
	p = copy_var_name(name);
	hash_put_str(env_ht_hidden, p, NULL);
}

int env_is_encrypted(const char *name) {
	if (NULL == env_ht_hidden) {
		char *p = copy_var_name(name);
		int ret = hash_contains_key_str(env_ht_encrypted, p);
		free(p);
		return ret;
	}
	return 0;
}

void env_set_encrypted(const char *name, const char *value) {
	if (NULL == env_ht_encrypted) {
		if (NULL == env_hash_ctx)
			env_hash_ctx = hash_ctx_new();
		env_ht_encrypted = hash_new(env_hash_ctx, 11); 
	}
	hash_put_str(env_ht_encrypted, strdup(name), bf_decrypt(value));
}

/* Spits out data blocks that report on the locale status of the agent.
 * This is called automatically by BuildEnv.  Anything like ping that
 * wants this information displayed but does not call BuildEnv should
 * call this directly, instead.
 */
void env_report_locale(char *buf) {
	char *want = platform_getenv("LANG");
	const char *got = platform_setlocale(want);

	if (NULL != want) {
		if (NULL == got) {
			int err = errno;
			got = platform_setlocale(NULL);
			format_message(buf, bfCat(E_EXEC_NOLOCALE), want, err);
			DataBlock(buf, NULL);
		}
		platform_freeenv(want);
	}

	if (NULL != got) {
		format_message(buf, bfCat(E_EXEC_LOCALE), got);
		DataBlock(buf, NULL);
	}
}

/* Everything in the switch statement below needs to appear in this list. */
const char *SUPPORTED_ENV_ACTIONS = "APEHSCUN";

/* envs is a string "S:X:Y\nA:C:D\nP:E:F\nC:Z:\n"  */
void BuildEnv(char *envs, char **src) {
	const char *locale;
	char    *cmd, *var, *value, *p;
	char    *newValue = NULL, *curr, *orig;
	char    *action, **argv;
	char    *bf_rptr = NULL;
	char    *bf_sptr = NULL;
	int     len;

	if (src)
		*src = NULL;

	/* If we have a locale, then make sure the LANG setting starts out
	 * matching it.
	 */
	locale = platform_setlocale("");
	if (NULL != locale)
		platform_setenv("LANG", locale, 0);

	action = calloc(1, BFBUFSIZ);
	cmd = bf_strtok(envs, "\n", &bf_rptr, &bf_sptr);
	do {
		memset(action, 0, BFBUFSIZ);
		if (!(var = strchr(cmd, ':')))
			continue;
		*var++ = 0;
		if (!(value = strchr(var, ':')))
			continue;
		*value++ = 0;

		if (strncmp(var, "BF_", 3) == 0)
			value = CleanPath(value);

		if (src && strcmp(var, ".source") == 0) {
			char *s;
			/* create a list of pre-exec source commands */
			value = env_expand_var(orig = value, 0, 0);
			if (!*src) {
				len = 0;
				s = *src = calloc(1, strlen(value) + 2);
			} else {
				len = strlen(*src);
				*src = realloc(*src, len + strlen(value) + 2);
				s = &(*src)[len];
			}
			len = strlen(value);
			memcpy(s, value, len);
			s += len;
			*s++ = '\n';
			*s++ = '\0';
			if (verbose) {
				format_message(action, bfCat(M_ENV_SRC), value);
				DataBlock(action, NULL);
			}
			free(value);
			continue;
		}

		if (strncmp(value, ".date", 5) == 0) {
			for (curr = &value[5] ; *curr && isspace((int)(*curr)) ; curr++);
			if (*curr != '\0') {
				newValue = expandDate(curr);
				platform_setenv(var, newValue, 1);
				format_message(action, bfCat(M_ENV_SET_INDIRECT),
						var, newValue, ".date", curr);
				free(newValue);
			} else {
				format_message(action, bfCat(E_ENV_CMD_NO_FORMAT), ".date");
			}
			if (verbose)
				DataBlock(action, NULL);
			continue;
		}

		if (strncmp(value, ".gmdate", 7) == 0) {
			for (curr = &value[7] ; *curr && isspace((int)(*curr)) ; curr++);
			if (*curr == '\0') {
				format_message(action, bfCat(E_ENV_CMD_NO_FORMAT), ".gmdate");
			} else {
				newValue = expandGMDate(curr);
				platform_setenv(var, newValue, 1);
				format_message(action, bfCat(M_ENV_SET_INDIRECT),
						var, newValue, ".gmdate", curr);
				free(newValue);
			}
			if (verbose)
				DataBlock(action, NULL);
			continue;
		}

		if (strncmp(value, ".substr", 7) == 0) {
			/* .substr "$STR" start [end] */
			argv = getArgs(&value[7]);
			if (!argv) {
				format_message(action, bfCat(E_ENV_SUBSTR_NO_ARGS));
			} else if (!argv[0]) {
				format_message(action, bfCat(E_ENV_SUBSTR_NO_SOURCE));
			} else if ( !argv[1] || !isdigit((int)(*argv[1])) ) {
				format_message(action, bfCat(E_ENV_SUBSTR_BAD_START_INDEX));
			} else if ( argv[2] && !isdigit((int)(*argv[2])) ) {
				format_message(action, bfCat(E_ENV_SUBSTR_BAD_END_INDEX));
			} else {
				argv[0] = curr = env_expand_var(orig = argv[0], 0, 0);
				newValue = strdup(curr);

				if ((int) atoi(argv[1]) < (int)strlen(argv[0]))
					strcpy(curr, &curr[atoi(argv[1])]);

				if (argv[2] && (int)atoi(argv[2]) <= (int)strlen(argv[0]))
					curr[atoi(argv[2])] = 0;
				format_message(action, bfCat(M_ENV_SUBSTR),
						var, argv[0], newValue, atoi(argv[1]),
						(argv[2]) ? atoi(argv[2]) : 0);
				platform_setenv(var, argv[0], 1);
				free(newValue);
				free(curr);
				argv[0] = orig; /* restore the original argv[0] pointer for freeArgs() */
			}
			freeArgs(argv);
			if (verbose)
				DataBlock(action, NULL);
			continue;
		}

		if (strncmp(value, ".regex", 6) == 0) {
			/* .regex /pattern/repl/ $value */
			/* .regex /pattern/repl/ value */
			regex_t *srch = NULL;
			char *spat = NULL;
			char *repl = NULL;
			char *target = NULL;

			curr = strdup(orig = &value[6]);
			while ( orig && isspace((int)(*orig)) )
				orig++;

			if (bf_regex_prep(&srch, &spat, &repl, &target, orig) != 0)
				format_message(action, bfCat(E_ENV_REGEX_BAD), target);
			else {
				newValue = bf_regex_apply(target, srch, repl);
				bf_regex_free(srch);
				format_message(action, bfCat(M_ENV_REGEX), var, newValue, spat, *spat, repl, *spat, target);
				platform_setenv(var, newValue, 1);
			}

			free(curr);
			free(spat);
			free(repl);
			free(target);
			free(newValue);
			if (verbose)
				DataBlock(action, NULL);
			continue;
		}

		/*
		 * Everything in the switch statement below needs to appear in
		 * SUPPORTED_ENV_ACTIONS.  Otherwise, they will get discarded
		 * when they are received.
		 */
		switch (*cmd) {
			case 'A':		/* append value to var */
			case 'P':		/* prepend value to var */
				value = env_expand_var(orig = value, 0, 0);
				/* convert a windows path to unix format */
				if (strcmp(var, "PATH") == 0 && platform_style() == STYLE_UNIX) {
					p = value;
					while (NULL != (p = strchr(p, ';')))
						*p++ = ':';
				}
				if (NULL != (curr = platform_getenv(var))) {
					newValue = calloc(1, strlen(value) + strlen(curr) + 2);
					if (*cmd == 'A') {
						format_message(newValue, "%s%c%s", curr, EnvPathSeparator(), value);
						format_message(action, bfCat(M_ENV_APPEND_PREPEND), var, curr, orig);
					}
					else {
						format_message(newValue, "%s%c%s", value, EnvPathSeparator(), curr);
						format_message(action, bfCat(M_ENV_APPEND_PREPEND), var, orig, curr);
					}
					platform_setenv(var, newValue, 1);
					free(newValue);
					platform_freeenv(curr);
				}
				else {
					platform_setenv(var, value, 1);
					format_message(action, bfCat(M_ENV_SET), var, orig);
				}
				free(value);
				break;
			case 'E':		/* set var to encrypted value */
				env_set_encrypted(var, value);
				format_message(action, bfCat(M_ENV_ENCRYPTED), var);
				break;
			case 'H':		/* set var to hidden value */
				value = env_expand_var(orig = value, 0, 0);
				platform_setenv(var, value, 1);
				env_set_hidden(var);
				if (strcmp(var, "_MAP") == 0) {
					if (mapStr)
						free(mapStr);
					mapStr = strdup(value);
				}
				format_message(action, bfCat(M_ENV_HIDDEN), var);
				free(value);
				break;
			case 'S':		/* set var to value */
				value = env_expand_var(orig = value, 0, 0);
				/* convert a windows path to unix format */
				if (strcmp(var, "PATH") == 0 && platform_style() == STYLE_UNIX) {
					p = value;
					while (NULL != (p = strchr(p, ';')))
						*p++ = ':';
				}
				if (strcmp(var, "_MAP") == 0) {
					if (mapStr)
						free(mapStr);
					mapStr = strdup(value);
					format_message(action, bfCat(M_ENV_SET), var, value);
				}
				else if (strcmp(var, "_TRAP") == 0) {
					if (trapCommand)
						free(trapCommand);
					trapCommand = strdup(value);
					format_message(action, bfCat(M_ENV_SET), var, value);
				}
				else if (strcmp(var, "_BFAGENTDEBUG") == 0) {
					if (!bf_debug) {
						bf_debug = calloc(1, BFBUFSIZ);
						strcpy(action, bfCat(M_ENV_DEBUG));
					}
				}
				else {
					platform_setenv(var, value, 1);
					format_message(action, bfCat(M_ENV_SET), var, orig);
				}
				free(value);
				break;
			case 'C':		/* clear the var */
				format_message(action, bfCat(M_ENV_CLEAR), var);
				platform_setenv(var, "", 1);
				break;
			case 'U':		/* unset the var - no way to know if this works on all platforms. */
				platform_unsetenv(var);
				format_message(action, bfCat(M_ENV_DELETE), var);
				break;
			case 'N':		/* set if not already set */
				if (NULL != (curr = platform_getenv(var))) {
					format_message(action, bfCat(M_ENV_SKIPPED_ALREADY_SET), var, curr, value);
					platform_freeenv(curr);
				}
				else {
					value = env_expand_var(value, 0, 0);
					platform_setenv(var, value, 0);
					format_message(action, bfCat(M_ENV_SET_NOT_ALREADY_SET), var, value);
					free(value);
				}
				break;
		}

		if (*action != '\0' && verbose)
			DataBlock(action, NULL);
	} while (NULL != (cmd = bf_strtok(NULL, "\n", &bf_rptr, &bf_sptr)));

	bf_strtok(NULL, NULL, NULL, &bf_sptr);

	/* set the agent version and OS platform */
	platform_setenv("BF_AGENT_VERSION", AGENT_VERSION, 1);
	format_message(action, bfCat(M_ENV_SET), "BF_AGENT_VERSION", AGENT_VERSION);
	DataBlock(action, NULL);

	{
		const char *id = platform_id();
		platform_setenv("BF_AGENT_PLATFORM", id, 1);
		format_message(action, bfCat(M_ENV_SET), "BF_AGENT_PLATFORM", id);
		DataBlock(action, NULL);
	}

	/* Pay attention to and report a change to the locale. */
	env_report_locale(action);

	cc_audit = 0;
	if (NULL != (p = platform_getenv("CLEARCASE_AUDIT"))) {
		cc_audit = 1;
		platform_freeenv(p);
	}
	if (!BFCONF_EXISTS(ignore_clearcase_view))
		clearcase_view = platform_getenv("CLEARCASE_VIEW");
	free(action);
}


int checkEnvRecursion(char *var, int len) {
	char *p;
	int *pCount;

	/* initial call allocs a cache */
	if (NULL == env_ht_recur) {
		if (NULL == env_hash_ctx)
			env_hash_ctx = hash_ctx_new();
		env_ht_recur = hash_new(env_hash_ctx, 89);
	}

	/* flush the cache? */
	if (var == NULL) {
		hash_clear(env_ht_recur);
		return 0;
	}

	/* look for it in the cache. if it's there, return the number of times we've seen it before. */
	if (NULL != (pCount = hash_get(env_ht_recur, var, len)))
		return ++*pCount;

	/* add the var, we haven't seen it yet */
	p = malloc(len);
	memcpy(p, var, len);
	pCount = calloc(1, sizeof(int));
	hash_put(env_ht_recur, p, len, pCount);
	return 0;
}

/*
 * expand a string that has an environment variable in it.
 *
 * handle translation of windows and unix for commonality
 *
 */

#define UA_NONE 0
#define UA_WARN 1
#define UA_USE_EMPTY 2

char *env_expand_var_impl(
		const char *src,
		int ignore_escapes,
		int do_encrypted,
		int do_hidden_check)
{
	char	*p, *value, *remainder, **env, *q;
	char	*result;
	int 	varlen, envlen, parse_skip, did_subst, len;
	int		default_undef_action = UA_WARN;
	int		undef_action;
	char	**env_ptr;
	char    *src2;
	char    *hidden_var;
	int		advance_pointer=0;

	if (!src)
		return NULL;

	/* clear the recursion cache */
	checkEnvRecursion(NULL, 0);

	/* Find out whether $VAR, ${VAR}, and %VAR% should expand as var name
	 * or in shell-compatible way.  For the time being, %VAR% will expand
	 * as blank instead of staying %VAR%, which is wrong...  That will be
	 * fixed later.
	 */
	if (BFCONF_EXISTS(shell_compatible_undef_vars))
		default_undef_action |= UA_USE_EMPTY;

	/* make a local copy of the source string */
	/* we use calloc/strcpy instead of strdup so we can realloc later */
	result = strdup(src);

	/* do windows env var conversion, %X% -> ${X} */
	for (p = result ; *p ; p++) {
		/* start of a var or just another character */
		if (*p != '%')
			continue;

		/* the % must preceed a valid variable */
		/* if it doesn't, skip over it and continue looking */
		if ((varlen = strspn(p+1, valid_var_chars)) < 1)
			continue;

		/* the terminating character has to be a % for this to be valid */
		if (*(p + varlen + 1) != '%')
			continue;

		/* it's a legal var, convert it to unix style */
		*p++ = 0;
		*(p + varlen) = '}';

		/* need to grow the string by one char, to insert the extra curly */
		src2 = strdup(p);
		parse_skip = strlen(result);
		result = realloc(result, parse_skip + strlen(src2) + 3);
		result[parse_skip++] = '$';
		result[parse_skip++] = '{';
		strcpy(&result[parse_skip], src2);
		free(src2);
		p = &result[parse_skip];
	}

	/* scan the result buffer for $ vars */
	/* p may get bumped off the end of the buffer so we check to see if that's the case */
	/* instead of checking *p as our for loop test. a segv safe test :) */
	parse_skip = 0;	/* how much to skip over on parser restarts */
	env_ptr = NULL;
	for (p = result ; p < &(result[strlen(result)]); ) {
		if (advance_pointer == 1) {
			p++;
		} else {
			advance_pointer = 1;
		}

		/* a single quote? */
		if (*p == '\'' && (q = strchr(p+1, '\'')) != NULL)
			p = q+1;

		/* an escaped character? */
		if (!ignore_escapes && *p == '\\' && platform_style() != STYLE_WINDOWS) {
			memmove(p, p+1, strlen(p));
			continue;
		}

		/* start of a variable ? */
		if (*p != '$')
			continue;

		/* if we have a $$, convert it to a single $ and skip over it */
		if (*(p+1) == '$') {
			memmove(p, p+1, strlen(p));
			continue;
		}

		/* at this point, p points to a $ */
		undef_action = default_undef_action;

		/* ${VAR} format */
		if (*(p+1) == '{') {
			/* is there a matching right curly after a valid variable name ? */
			if (0 != (varlen = strspn(p+2, valid_var_chars)) && p[varlen + 2] == '}') {
				/* Skip over the ${ */
				memmove(p, p+2, varlen);
				/* Eat the trailing } */
				memmove(p + varlen, p + varlen + 3, strlen(p + varlen + 3) + 1);
			}
			else {
				++p;
				continue;
			}
		}

		/* $[VAR] is like ${VAR}, but if a variable is undefined, it will always
		 * use "" for the value instead of "VAR", and it does not generate the
		 * warning message.
		 */
		else if (*(p+1) == '[') {
			undef_action = UA_USE_EMPTY;
			/* is there a matching right bracket after a valid variable name ? */
			if ((varlen = strspn(p+2, valid_var_chars)) && *(p + varlen + 2) == ']') {
				/* Skip over the $[ */
				memmove(p, p+2, varlen);
				/* Eat the trailing ] */
				memmove(p + varlen, p + varlen + 3, strlen(p + varlen + 3) + 1);
			}
			else {
				++p;
				continue;
			}
		}

		/* $VAR format */
		else {
			if ((varlen = strspn(p+1, valid_var_chars))) {	/* find the length of this var */
				/* it's legal, clean it up - this eats the $ */
				memmove(p, p+1, strlen(p));
			}
			else {
				/* not a legal variable, skip over it */
				++p;
				continue;
			}
		}


		/* p points to a variable name and varlen says how long it is */
		parse_skip = p - result;	/* remember how much is already done to protect expanded $'s */
		if (env_recursion_limit && checkEnvRecursion(p, varlen) > env_recursion_limit) {
			q = calloc(1, BFBUFSIZ);
			strcpy(q, bfCat(E_ENV_RECURSION_SUPPRESSED));
			strncat(q, p, varlen);
			DataBlock(q, NULL);
			free(q);
		} else {
			if (do_hidden_check && NULL != env_ht_hidden) {
				hidden_var = calloc(1, varlen+1);
				strncpy(hidden_var, p, varlen);
				hidden_var[varlen] = 0;

				if (env_is_hidden(hidden_var)) {
					src2 = calloc(1, BFBUFSIZ);
					format_message(src2, bfCat(E_ENV_HIDDEN), hidden_var);
					DataBlock(src2, NULL);
					free(src2);
					free(hidden_var);

					free(result);
					result = strdup(src);
					if (env_ptr != NULL)
						freeBFenviron(env_ptr);
					return result;
				}
				free(hidden_var);
			}

			did_subst = 0;
			for (env = env_ptr = getBFenviron() ; *env ; env++) {
				/* get the length of the env variable */
				envlen = strcspn(*env, "=");

				/* compare the variable name pointed to by p with the env variable name */
				if (envlen == varlen && BFStrCmp(p, *env, varlen) == 0) {
					value = strchr(*env, '=');
					value++;

					/* null the working string and point to the remainder after the variable */
					*p = 0;
					remainder = strdup(p + varlen);	/* make a copy of the stuff after the var name in the working string */
												/* because we're going to realloc the working string buffer shortly */
					/* rebuild the working string */
					len = strlen(result);
					result = realloc(result, len + strlen(value) + strlen(remainder) + 1);
					strcpy(&result[len], value);
					strcat(&result[len], remainder);
					free(remainder);
					p = result + parse_skip;
					advance_pointer=0;
					did_subst = 1;
					break;
				}
			}

			if (!did_subst && do_encrypted && NULL != env_ht_encrypted) {
				hash_iter *hi = hash_iter_new(env_ht_encrypted);
				while (hash_iter_next(hi)) {
					q = hi->entry->key;
					value = hi->entry->value;
					if (strcmp(q, p) == 0) { /* a match! */
						/* null the working string and point to the remainder after the variable */
						*p = 0;
						remainder = strdup(p + varlen);	/* make a copy of the stuff after the var name in the working string */
														/* because we're going to realloc the working string buffer shortly */
						/* rebuild the working string */
						len = strlen(result);
						result = realloc(result, len + strlen(value) + strlen(remainder) + 1);
						strcpy(&result[len], value);
						strcat(&result[len], remainder);
						free(remainder);
						p = result + parse_skip;
						advance_pointer = 0;
						did_subst = 1;
						break;
					}
				}
				hash_iter_destroy(hi);
			}

			if (!did_subst) {
				/* Not $[VAR] format, so warn about undefined var */
				if (undef_action & UA_WARN) {
					q = calloc(1, BFBUFSIZ);
					strcpy(q, bfCat(E_ENV_UNDEF));
					strncat(q, p, varlen);
					DataBlock(q, NULL);
					free(q);
				}

				/* $[VAR] format or shell_compatible_undef_vars was set */
				if (undef_action & UA_USE_EMPTY)
					strcpy(p, (p+varlen));
			}

		}

	}
	freeBFenviron(env_ptr);
	return result;
}


