/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

#ifdef IO_DEBUG
# if IO_DEBUG > 1
#  define D(x) printf("%s\n", x)
# else
#  define D(x) DebugBlock(x)
# endif
#endif

extern int disable_telnet_support;

#ifdef HAVE_SSL

extern int ifd;
extern int ofd;

void	*ssl = NULL;
SSL_CTX	*ctx = NULL;

int verify_SSL_callback(int flag, X509_STORE_CTX *store) {
	char	*buf;
	int 	err;

	if (!flag) {
		err = X509_STORE_CTX_get_error(store);
		buf = calloc(1, BFBUFSIZ);
		format_message(buf, bfCat(E_SSL_UNVERIFIED_CERT), err, X509_verify_cert_error_string(err));
		return 0;
	}
	return 1;
}

int init_CTX(void) {
	char	*cert, *ca;
	int 	rc;

	if (ctx)
		return SE_OK;

	rc = SE_OK;
	agent_mode = SE_OK;

	/* if a server certificate is defined, then require SSL mode */
	if ((cert = bfconf_get("cert", NULL)) != NULL) {
		agent_mode = SE_OK_SECURE;
		if (!SSL_library_init())
			rc = SE_INIT;
		else {
			SSL_load_error_strings();
			ctx = SSL_CTX_new(TLSv1_server_method());
			SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
			SSL_CTX_set_options(ctx, SSL_OP_TLS_ROLLBACK_BUG);
			SSL_CTX_set_cipher_list(ctx, "DEFAULT:!MD5:!LOW:!EXP:@STRENGTH");
		}

		if (rc == SE_OK && SSL_CTX_use_certificate_file(ctx, cert, SSL_FILETYPE_PEM) != 1)
			rc = SE_BADCERT;

		if (rc == SE_OK && SSL_CTX_use_PrivateKey_file(ctx, cert, SSL_FILETYPE_PEM) != 1)
			rc = SE_BADPKEYFILE;

		if (rc == SE_OK && SSL_CTX_check_private_key(ctx) != 1)
			rc = SE_BADPKEY;

		if (rc == SE_OK && (ca = bfconf_get("ca", NULL)) != NULL) {
			if (SSL_CTX_load_verify_locations(ctx, ca, NULL) != 1)
				rc = SE_BADCA;
			else {
				SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, verify_SSL_callback);
				agent_mode = SE_OK_CERTIFIED;
			}
			free(ca);
		}
		free(cert);
		if (rc != SE_OK) {
			SSL_CTX_free(ctx);
			ctx = NULL;
		}
	}
	return rc;
}

void stop_SSL(void) {
	sayHello(CODE_GOODBYE);
	if (ssl) {
		SSL_shutdown(ssl);
		SSL_free(ssl);
		ssl = NULL;
	}
	close(0);
	close(1);
	close(2);
	close(ifd);
	close(ofd);
#ifdef WIN32
	if (win_sock) {
		shutdown(ifd, 0);
		closesocket(ifd);
		WSACleanup();
	}
#endif
}

void err_SSL(int rc) {
	switch (rc) {
		case SE_INIT:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_INIT));
			break;
		case SE_BIOIN:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BIOIN));
			break;
		case SE_BIOOUT:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BIOOUT));
			break;
		case SE_BADCA:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BADCA));
			break;
		case SE_BADCERT:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BADCERT));
			break;
		case SE_BADPKEYFILE:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BADPKEYFILE));
			break;
		case SE_BADPKEY:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_BADPKEY));
			break;
		case SE_NOSSL:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_NOSSL));
			break;
		case SE_HANDSHAKE:
			ResultBlock(rc, -1, bfCat(E_SSL_SE_HANDSHAKE));
			break;
		default:
			ResultBlock(rc, -1, bfCat(E_SSL_UNKNOWN));
			break;
	}
}

void *start_SSL(void) {
	int 	rc;

	if ((rc = init_CTX()) != SE_OK) {
		err_SSL(rc);
		return NULL;
	}

	switch (agent_mode) {
		case SE_OK:
			sayHello(CODE_HELLO);
			return NULL; /* not null, but not an SSL pointer */
		case SE_OK_SECURE:
			sayHello(CODE_SECURE_HELLO);
			break;
		case SE_OK_CERTIFIED:
			sayHello(CODE_CERTIFIED_HELLO);
			break;
	}

	if ((ssl = SSL_new(ctx)) == NULL) {
		err_SSL(SE_NOSSL);
		return NULL;
	}

	SSL_set_fd(ssl, ifd);

	SSL_set_accept_state(ssl);
	if (SSL_do_handshake(ssl) < 1) {
		SSL_free(ssl);
		ssl = NULL;
		err_SSL(SE_HANDSHAKE);
		return NULL;
	}
	return ssl;
}


void report_SSL(void) {
	char	*buf, *line;
	X509	*cert;

	if (!ssl) {
		DataBlock(bfCat(M_SSL_PLAINTEXT), NULL);
		return;
	}
	buf = calloc(1, BFBUFSIZ);

	format_message(buf, bfCat(M_SSL_STARTTLS), SSL_get_cipher_name(ssl));
	DataBlock(buf, NULL);

	if (agent_mode != SE_OK_CERTIFIED) {
		format_message(buf, bfCat(E_SSL_UNVERIFIED_CERT));
		DataBlock(buf, NULL);
	}

    if ((cert = SSL_get_peer_certificate(ssl))) {	/* Get certificates (if available) */
        if (NULL != (line = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0))) {
	        format_message(buf, bfCat(M_SSL_CONSOLE_ISSUER), line);
			DataBlock(buf, NULL);
			free(line);
		}
        if (NULL != (line = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0))) {
	        format_message(buf, bfCat(M_SSL_CONSOLE_CERT), line);
			DataBlock(buf, NULL);
			free(line);
		}
		X509_free(cert);
    }
	free(buf);
}

#define io_read(buf, len) (ssl ? SSL_read(ssl, buf, len) : platform_protocol_read(buf, len))
#define io_write(buf, len) (ssl ? (SSL_write(ssl, buf, len) >= 0) : platform_protocol_write(buf, len))

#else /* !HAVE_SSL */

void *start_SSL(void) {
	sayHello(CODE_HELLO);
	return NULL;
}

void stop_SSL(void) {
	sayHello(CODE_GOODBYE);
}

#define io_read(buf, len) platform_protocol_read(buf, len)
#define io_write(buf, len) platform_protocol_write(buf, len)

#endif /* HAVE_SSL */



#define PUTS_BUF_SIZE (BFBUFSIZ << 2)
static char *puts_buf = NULL;


int io_puts(const char *s) {
	int len = strlen(s);
	int ok;

	/* Writing UTF-8 to a socket ... */
	if (platform_protocol_has_socket() ||
			BFCONF_EXISTS(disable_transcode))
	{
#ifdef ZOS
		/* On z/OS, our UTF-8 data is transcoded to EBCDIC, and we
		 * need to trancode it back to real UTF-8 before we send it.
		 */
		int ret;
		char *p = malloc(len);
		memcpy(p, s, len);
		__etoa_l(p, len);
		ret = io_write(p, len);
		free(p);
		return ret;
#else /* !ZOS */
		/* On other platforms, UTF-8 is just UTF-8. */
		return io_write(s, len);
#endif /* ZOS */
	}

	/* Writing to the OS, so use its console encoding ... */
	if (platform_utf8_is_console()) {
		ok = io_write(s, len);
	} else {
		if (NULL == puts_buf)
			puts_buf = malloc(PUTS_BUF_SIZE);
		len = convert_utf8_to_console(s, len, puts_buf, PUTS_BUF_SIZE);
		ok = io_write(puts_buf, len);
	}
	return ok;
}

#define TELNET_SE   (240)
#define TELNET_NOP  (241)
#define TELNET_MARK (242)
#define TELNET_BRK  (243)
#define TELNET_IP   (244)
#define TELNET_AO   (245)
#define TELNET_AYT  (246)
#define TELNET_EC   (247)
#define TELNET_EL   (248)
#define TELNET_GA   (249)
#define TELNET_SB   (250)
#define TELNET_WILL (251)
#define TELNET_WONT (252)
#define TELNET_DO   (253)
#define TELNET_DONT (254)
#define TELNET_IAC  (255)

#define TNO_BINARY      (0)
#define TNO_SUPPRESS_GA (3)
#define TNO_TIMING_MARK (6)

#define GETS_UTF8_SIZE (BFBUFSIZ << 1)
static char gets_utf8[GETS_UTF8_SIZE];
static int gets_utf8_len = 0;

typedef enum {
	TNS_RESET,
	TNS_IAC,
	TNS_WILL,
	TNS_WONT,
	TNS_DO,
	TNS_DONT
} telnet_state_t;

static telnet_state_t telnet_state = TNS_RESET;
static const uint8_t TELNET_RESPONSE_AYT[] = {
	/* "\r\n[IBM Rational Build Forge Agent]\r\n" in ASCII */
	0x0D, 0x0A, 0x5B,
	0x49, 0x42, 0x4D, 0x20,
	0x52, 0x61, 0x74, 0x69, 0x6F, 0x6E, 0x61, 0x6C, 0x20,
	0x42, 0x75, 0x69, 0x6C, 0x64, 0x20,
	0x46, 0x6F, 0x72, 0x67, 0x65, 0x20,
	0x41, 0x67, 0x65, 0x6E, 0x74,
	0x5D, 0x0D, 0x0A
};

int io_ayt(void) {
	return io_write(TELNET_RESPONSE_AYT, sizeof(TELNET_RESPONSE_AYT));
}

static void io_gets_telnet_option(uint8_t code) {
	uint8_t buf[3];
	buf[0] = TELNET_IAC;
	/* buf[1] = response */
	buf[2] = code;

	switch (code) {
		/* Options we will claim to support, even though we don't actually
		 * do anything special for them.
		 */
		case TNO_BINARY:
		case TNO_SUPPRESS_GA:
		case TNO_TIMING_MARK:
		{
			switch (telnet_state) {
				case TNS_WILL: buf[1] = TELNET_DO;   break;
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   buf[1] = TELNET_WILL; break;
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;  
			}
			break;
		}

		/* We don't even understand what they're asking us to do,
		 * so refuse to cooperate.
		 */
		default:
		{
			switch (telnet_state) {
				case TNS_WILL: /* FALL THROUGH */
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   /* FALL THROUGH */
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;
			}
		}
	}

	io_write(buf, 3);
}

/* At some point, we discovered a 0xFF byte in the protocol stream.  That
 * is not a valid byte in UTF-8, so we will assume that a user is connected
 * to the agent using telnet and that it came from the telnet client.  This
 * is annoying, but we'll deal with it as best as we can by responding
 * appropriately when we see telnet control sequences.  If we encounter an
 * incomplete sequence, we'll save it for later.  Some sequences can be
 * quite long, so we give up if it looks like a sequence is too long to
 * bother trying to honor it.  In this case, we discard the IAC and
 * release the rest of the data to be processed as (alleged) UTF-8.
 */
static void io_gets_telnet(char *telnet_mark) {
	char *p = telnet_mark;
	char *pend = &gets_utf8[gets_utf8_len];

	/* Miniature FSM for telnet option negotiation */
	while (1) {
		/* Exhausted the buffer.  Truncate it at the telnet sequence. */
		if (p >= pend) {
			gets_utf8_len = telnet_mark - gets_utf8;
			return;
		}

		switch (telnet_state) {
			case TNS_RESET:
			{
				p = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
				if (NULL == p)
					return;
				++p;
				telnet_state = TNS_IAC;
				continue;
			}

			case TNS_IAC:
			{
				switch ((uint8_t)(*p++)) {
					case TELNET_AYT:
						io_ayt();
						goto completed;
					case TELNET_EC:
						if (telnet_mark > gets_utf8)
							--telnet_mark;
						goto completed;
					case TELNET_EL:
						telnet_mark = memrchr(gets_utf8, 0x0A, gets_utf8_len);
						if (NULL == telnet_mark)
							telnet_mark = gets_utf8;
						goto completed;
					case TELNET_WILL:
						telnet_state = TNS_WILL;
						continue;
					case TELNET_WONT:
						telnet_state = TNS_WONT;
						continue;
					case TELNET_DO:
						telnet_state = TNS_DO;
						continue;
					case TELNET_DONT:
						telnet_state = TNS_DONT;
						continue;
					case TELNET_IAC:
						++telnet_mark;
						goto completed;
					default:
						goto completed;
				}
			}

			default: /* WILL WONT DO DONT */
				io_gets_telnet_option((uint8_t)*p++);
				goto completed;
		}

completed:
		gets_utf8_len -= p - telnet_mark;
		pend = &gets_utf8[gets_utf8_len];
		telnet_state = TNS_RESET;
		if (telnet_mark < &gets_utf8[gets_utf8_len]) {
			memmove(telnet_mark, p, pend - telnet_mark);
			p = telnet_mark = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
			if (NULL != telnet_mark)
				continue;
		}
		return;
	}
}

/* Read UTF-8 data directly into the buffer ... */
static int io_gets_fill_utf8(void) { while (1) {
	int len;
	char *p = &gets_utf8[gets_utf8_len];
	char *q = p;

	len = io_read(p, GETS_UTF8_SIZE - gets_utf8_len);
#ifdef IO_DEBUG
	if (bf_debug) {
		format_message(bf_debug, "io_gets_fill_utf8: io_read -> len=%d\n", len);
		D(bf_debug);
	}
#endif /* IO_DEBUG */

	if (len <= 0) { return len; }

	p = &gets_utf8[gets_utf8_len];
	gets_utf8_len += len;

	/* If this is a person using a telnet, then we might see telnet
	 * control sequences.  This deals with them if there are any.
	 */
	if (!disable_telnet_support) {
		if (telnet_state != TNS_RESET || NULL != (q = memchr(p, 0xFF, len))) {
			io_gets_telnet(q);
			len = &gets_utf8[gets_utf8_len] - p;
			if (len <= 0)
				continue;
		}
	}

#ifdef ZOS
	__atoe_l(p, len);
#endif
	return gets_utf8_len;
}}


#define GETS_CONSOLE_SIZE (BFBUFSIZ)
/* Read from the OS, using its console encoding ... */
static char *gets_console = NULL;
static int gets_console_len = 0;

/* Reads from the console until there is a suitable boundary for
 * transcoding.  If the buffer fills completely and there is no
 * candidate boundary, then return the entire buffer.
 */
static int io_gets_fill_console(void) { while (1) {
	int len = io_read(&gets_console[gets_console_len],
			GETS_CONSOLE_SIZE - gets_console_len);
#ifdef IO_DEBUG
	if (bf_debug) {
		format_message(bf_debug, "io_gets_fill_console: io_read -> len=%d\n", len);
		D(bf_debug);
	}
#endif /* IO_DEBUG */

	/* The read returned an error or EOF.  If we have data buffered,
	 * then we flush it.  Otherwise, pass on the error/EOF result.
	 */
	if (len <= 0)
		return (gets_console_len > 0) ? gets_console_len : len;
	gets_console_len += len;

	/* See if we can find a good stopping point for transcoding. */
	len = convert_console_find_boundary(gets_console, gets_console_len);
#ifdef IO_DEBUG
	if (bf_debug) {
		format_message(bf_debug,
				"io_gets_fill_console: "
				"convert_console_find_boundary: %d -> %d\n",
				gets_console_len, len);
		D(bf_debug);
	}
#endif /* IO_DEBUG */
	if (len > 0)
		return len;

	/* If the buffer is full, give up and transcode the whole thing. */
	if (gets_console_len >= GETS_CONSOLE_SIZE) {
#ifdef IO_DEBUG
		if (bf_debug)
			D("io_gets_fill: console buffer overflow");
#endif /* IO_DEBUG */
		return GETS_CONSOLE_SIZE;
	}

	/* Not at a stopping point and still have room, so keep going... */
}}




static int io_gets_fill_transcode(void) {
	int len, converted;

	len = io_gets_fill_console();
	if (len <= 0) {
		if (gets_console_len <= 0)
			return len;
		len = gets_console_len;
	}

	/* Convert a block of console text to UTF-8. */
	converted = convert_console_to_utf8(gets_console, len,
			&gets_utf8[gets_utf8_len], GETS_UTF8_SIZE - gets_utf8_len);

#ifdef IO_DEBUG
	if (bf_debug) {
		format_message(bf_debug,
				"io_gets_fill_transcode: console [%d - %d] -> UTF8 [%d + %d]",
				gets_console_len, len,
				gets_utf8_len, converted);
		D(bf_debug);
	}
#endif /* IO_DEBUG */

	gets_console_len -= len;
	memmove(gets_console, &gets_console[len], gets_console_len);
	return gets_utf8_len += converted;
}

static int (*io_gets_fill_fn)(void) = NULL;
#define io_gets_fill() ((*io_gets_fill_fn)())

char *io_gets(char *buf, int maxlen) {
	int i=0;

	/* figure out if we are going to fill the buffer using a
	 * network socket or the console.
	 */
	if (NULL == io_gets_fill_fn) {
		if (platform_protocol_has_socket() ||
				BFCONF_EXISTS(disable_transcode)) {
			io_gets_fill_fn = &io_gets_fill_utf8;
		} else {
			io_gets_fill_fn = &io_gets_fill_transcode;
			gets_console = malloc(GETS_CONSOLE_SIZE);
		}
	}

	/* If we fill the buffer, then stop regardless of whether or not
	 * we find a newline.
	 */
	while (i < maxlen-1) {
		/* If we need more chars than are currently in the UTF-8 buffer,
		 * then we need to fill it.  If we can't, then return what we
		 * have, or NULL if we don't have anything.
		 */
		while (i >= gets_utf8_len) {
			if (io_gets_fill() <= 0) {
				if (gets_utf8_len <= 0) {
#ifdef IO_DEBUG
					if (bf_debug)
						D("io_gets: NULL");
#endif /* IO_DEBUG */
					return NULL;
				}
				break;
			}
		}

		/* Non-control character -- preserve and continue */
		if (gets_utf8[i] & ~0x1F) {
			++i;
			continue;
		}

		/* Tolerate tab chars, too */
		if (gets_utf8[i] == '\t') {
			++i;
			continue;
		}

		/* If it's \n, we have a full line. */
		if (gets_utf8[i] == '\n') {
			++i;
			break;
		}

		/* Backspace?  You must be using the broken Windows telnet...  We
		 * only honor it if there is something to delete; otherwise, we
		 * just discard it.
		 */
		if (gets_utf8[i] == '\b') {
			if (i > 0) {
				--i;
				gets_utf8_len -= 2;
				memmove(&gets_utf8[i], &gets_utf8[i+2], gets_utf8_len - i);
			} else {
				--gets_utf8_len;
				memmove(&gets_utf8[i], &gets_utf8[i+1], gets_utf8_len - i);
			}
			continue;
		}

		if (gets_utf8[i] == '\r') {
			--gets_utf8_len;
			memmove(&gets_utf8[i], &gets_utf8[i+1], gets_utf8_len - i);
			continue;
		}

		/* Any other control char -- immediately hang up on the client */
		printf("Unexpected char: 0x%02x\n", (unsigned int)gets_utf8[i]);
		io_ayt();
		return NULL;
	}

	/* Return the data to the caller and remove it from the buffer. */
	memcpy(buf, gets_utf8, i);
	buf[i] = 0;
	gets_utf8_len -= i;
	memmove(gets_utf8, &gets_utf8[i], gets_utf8_len);

	/* Don't trust the client to speak UTF-8 properly... Some things, like
	 * overlong encoding, can be dangerous if they are allowed.
	 */
	convert_utf8_clean(buf, i, 1);

#ifdef IO_DEBUG
	if (bf_debug) {
		format_message(bf_debug, "io_gets: [%d]=\"%s\"", i, buf);
		D(bf_debug);
	}
#endif /* IO_DEBUG */

	return buf;
}


