/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* On z/OS, these real UTF-8 strings need to be transcoded into EBCDIC before
 * we can return them.  This presents a problem due to memory-management
 * concerns.  We deal with it by caching them in a hash table so that they
 * will stay allocated.  We clear the hash table if the language changes.
 */
#ifdef ZOS
static hash_ctx *ctx = NULL;
static hash_table *ht = NULL;
#endif

/*
to add a language, create a new _str_LANG[] pointer array.
the included language file MUST have strings indexed for every
code in codes.h. 
update bfSetLang to set the internal table pointer appropriately
for the language in question.
*/

static const char *_str_de[] = {
#include "lang_de.h"
NULL
};

static const char *_str_en[] = {
#include "lang_en.h"
NULL
};

static const char *_str_es[] = {
#include "lang_es.h"
NULL
};

static const char *_str_fr[] = {
#include "lang_fr.h"
NULL
};

static const char *_str_it[] = {
#include "lang_it.h"
NULL
};

static const char *_str_ja[] = {
#include "lang_ja.h"
NULL
};

static const char *_str_ko[] = {
#include "lang_ko.h"
NULL
};

static const char *_str_pt[] = {
#include "lang_pt.h"
NULL
};

static const char *_str_zh_CN[] = {
#include "lang_zh_CN.h"
NULL
};

static const char *_str_zh_TW[] = {
#include "lang_zh_TW.h"
NULL
};

/* string name of the currently selected language */
static char	      *lang = NULL;

/* pointer to the string table for the current language */
static const char  **curr_lang = NULL;

void bfSetLang(const char *newLang) {
	int fallback_lang = 1;
	int fallback_locale = 1;
	char *p;

	if (lang) {
		free(lang);
#ifdef ZOS
		if (ht != NULL)
			hash_clear(ht);
#endif
	}

	lang = strdup(newLang);

	while (1) {
		/* Look for a match to one of the supported languages. */
		if (strcmp(lang,"de") == 0) {
			curr_lang = _str_de;
			platform_set_language(LANG_GERMAN, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"en") == 0) {
			curr_lang = _str_en;
			platform_set_language(LANG_ENGLISH, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"es") == 0) {
			curr_lang = _str_es;
			platform_set_language(LANG_SPANISH, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"fr") == 0) {
			curr_lang = _str_fr;
			platform_set_language(LANG_FRENCH, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"it") == 0) {
			curr_lang = _str_it;
			platform_set_language(LANG_ITALIAN, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"ja") == 0) {
			curr_lang = _str_ja;
			platform_set_language(LANG_JAPANESE, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"ko") == 0) {
			curr_lang = _str_ko;
			platform_set_language(LANG_KOREAN, SUBLANG_DEFAULT);
		}
		else if (strcmp(lang,"pt") == 0) {
			curr_lang = _str_pt;
			platform_set_language(LANG_PORTUGUESE, SUBLANG_PORTUGUESE_BRAZILIAN);
		}
		else if (strcmp(lang,"zh") == 0) {
			curr_lang = _str_zh_CN;
			platform_set_language(LANG_CHINESE, SUBLANG_CHINESE_SIMPLIFIED);

		}
		else if (strcmp(lang,"zh_TW") == 0) {
			curr_lang = _str_zh_TW;
			platform_set_language(LANG_CHINESE, SUBLANG_CHINESE_TRADITIONAL);
		}

		/* Strip off the modifier, if there is one. */
		else if (NULL != (p = strrchr(lang, '@'))) {
			*p = '\0';
			continue;
		}

		/* Strip off the codeset, if there is one. */
		else if (NULL != (p = strrchr(lang, '.'))) {
			*p = '\0';
			continue;
		}

		/* Strip off the territory, if there is one. */
		else if (NULL != (p = strrchr(lang, '_'))) {
			*p = '\0';
			continue;
		}

		/* Try falling back on conf file's "lang" setting. */
		else if (fallback_lang) {
			fallback_lang = 0;
			free(lang);
			lang = bfconf_get("lang", "");
			continue;
		}

		/* Try falling back on the current locale, or the conf
		 * file setting if locales are not supported.
		 */
		else if (fallback_locale) {
			fallback_locale = 0;
			free(lang);
			{
				const char *locale = platform_setlocale(NULL);
				if (locale != NULL) {
					lang = strdup(locale);
					continue;
				}
			}
			lang = bfconf_get("locale", "en");
			continue;
		}

		/* Out of options, so just use English. */
		else {
			free(lang);
			lang = strdup("en");
			curr_lang = _str_en;
			platform_set_language(LANG_ENGLISH, SUBLANG_DEFAULT);
		}
		break;
	}
}




#ifdef ZOS
static unsigned int zos_cache_hash_fn(const void *unused, size_t idx)
{
	return idx;
}

static int zos_cache_equals_fn(
		const void *unused1, size_t idx1,
		const void *unused2, size_t idx2)
{
	return idx1 == idx2;
}

const char *bfCat(int code) {
	char *msg;

	if (lang == NULL)
		bfSetLang("");

	if (ctx == NULL) {
		ctx = hash_ctx_new();
		ctx->hash_fn = zos_cache_hash_fn;
		ctx->equals_fn = zos_cache_equals_fn;
		ctx->free_key_fn = hash_ignore_free_fn;
		ht = hash_new(ctx, 89);
	} else if (NULL != (msg = hash_get(ht, NULL, code))) {
		/* Found the msg in the existing cache */
		return msg;
	}

	/* Just now initializing, or did not find the msg. */
	msg = strdup(curr_lang[code]);
	__atoe_l(msg, strlen(msg)+1);
	hash_put(ht, NULL, code, msg);
	return msg;
}

#else /* !def ZOS */

const char *bfCat(int code) {
	if (lang == NULL)
		bfSetLang("");
	return curr_lang[code];
}

#endif

