/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include <fcntl.h>
#include "md5.h"

#ifdef WIN32
#include <direct.h>
#endif


int     datablock_cache_size = 0;
char *  datablock_cache = NULL;
int     datablock_cache_enabled = 0;

int     output_buffer_size = 0;
char *  output_buffer = NULL;

int     did_hello = 0;

#define CODE_HELLO_MSG           "HELLO - BuildForge Agent v" AGENT_ID_VERSION
#define CODE_SECURE_HELLO_MSG    "HELLO TLS - BuildForge Agent v" AGENT_ID_VERSION
#define CODE_CERTIFIED_HELLO_MSG "HELLO TLS - BuildForge Agent v" AGENT_ID_VERSION
#define CODE_GOODBYE_MSG         "GOODBYE - BuildForge Agent v" AGENT_ID_VERSION

extern FILE *  LogF;

#ifndef WIN32
extern int     errno;
#endif

/*
 * function to close an open logfile
 */
void CloseLog(void)
{
	if (LogF)
		fclose(LogF);
	LogF = NULL;
	MonitorClose();
}

/*
 * function to open a log file if one is specified in a request
 */
void OpenLog(void)
{
	char	*file, *msg;
	char	*root;

	CloseLog();
	MonitorOpen();
	if ((file = platform_getenv("BF_LOG")) == NULL)
		file = platform_getenv("_LOG");

	if (file && *file != '\0') {
		CleanPath(file);
		if (*file != PathSeparator() && file[1] != ':') {
			root = malloc(BFBUFSIZ);
			platform_getcwd(root, BFBUFSIZ);
			sprintf(&root[strlen(root)], "%c%s", PathSeparator(), file);
			platform_freeenv(file);
			file = root;
		}
		else {
			root = strdup(file);
			platform_freeenv(file);
			file = root;
		}
		CleanPath(file);

		msg = calloc(1, BFBUFSIZ);
		LogF = platform_fopen(file, "a");
		if (LogF != NULL) {
			setbuf(LogF, NULL);
			format_message(msg, bfCat(M_LOG_PATHNAME), file);
		}
		else {
			format_message(msg, bfCat(E_LOG_PATHNAME_FAIL), file);
		}
		DataBlock(msg, NULL);
		free(msg);
		free(file);
	}
}

/*
 * function to log a data block when logging is enabled
 */
void WriteLog(const char *msg)
{
	char	*s;

	if (msg) {
		s = strdup(msg);
		RemoveTrailingNewline(s);
		if (LogF)
			fprintf(LogF, "%s\n", s);
/*		MonitorOutput(s); */
		free(s);
	}
}

int FlushOutputCache(void) {
	int ok;

	if (*datablock_cache == '\0')
		return 1;
	ok = io_puts(datablock_cache); /* just need a true/false on this */
	memset(datablock_cache, 0, datablock_cache_size);
	return ok;
}

int OutputCache(cache_request_t mode) {
	if (!datablock_cache)
		return 1;

	switch (mode) {
		case CACHE_ON:
			datablock_cache_enabled = 1;
			break;
		case CACHE_OFF:
			datablock_cache_enabled = 0;
			/* FALL THROUGH */
		case CACHE_FLUSH:
			return FlushOutputCache();
	};
	return 1;
}

/*
 * function to set up an output buffer. start with BUFSIZ
 * and grow it if needed, and grow big.
 * we use an output buffer so we can easily tell when writes
 * to stdout fail, which means it's time to quit...
 *
 */
void FixOutputBuffer(int size)
{
	if (output_buffer_size > size)
		return;
	size += 4096;
	if (output_buffer)
		free(output_buffer);
	output_buffer = calloc(1, size);
	output_buffer_size = size;
}

/*
 *  function to output a result block 
 */
int ResultBlock(int code, int attr, const char *message)
{
	int len;
	FixOutputBuffer(strlen(message) + 30);

	len = format_message(output_buffer, "250 RESULT %d",
			code >= 0 ? code : -code);
	if (attr >= 0)
		len += format_message(&output_buffer[len], " %d", attr);
	output_buffer[len++] = '\n';
	if (message) {
		int msglen = strlen(message);
		memcpy(&output_buffer[len], message, msglen);
		len += msglen;
		output_buffer[len++] = '\n';
	}
	strcpy(&output_buffer[len], ".\n");
	WriteLog(message);

	/* dump the output cache if there's data in it */
	/* and switch it off */
	OutputCache(CACHE_OFF);

	if (!io_puts(output_buffer))
		return 0;
	io_puts(END_MESSAGE);
	return 1;
}

time_t beatmark = 0;
char	beatbuf[32];
int HeartBeat(void) {
	time_t now;

	time(&now);
	if (now - beatmark > heartbeat_interval) {
		OutputCache(CACHE_FLUSH);
		beatmark = now;
		format_message(beatbuf, "300 HEARTBEAT %d\n", bf_LoadValue());
		return io_puts(beatbuf);
	}
	return 1;
}

/* output a string through the cache mechanism */
int CachedOutput(const char *buf) {
	int 	len, clen;

	len = (int)strlen(buf);
	if (!datablock_cache || !datablock_cache_enabled)
		return io_puts(buf);
	clen = (int)strlen(datablock_cache);
	if (clen + len >= datablock_cache_size)	/* would overflow the cache */
		OutputCache(CACHE_FLUSH);
	if (len > datablock_cache_size)		/* won't fit in the cache at all */
		return io_puts(buf);
	strcat(datablock_cache, buf);
	return 1;
}

/*
 * subroutine to output a log data block.
 * dispatch routines can insert log data
 * into the normal stream in this manner.
 */
int DataBlock(const char *data, const char *info) {
	int datalen;
	if (NULL == data) {
		data = "";
		datalen = 0;
	} else {
		datalen = strlen(data);
	}

	if (info) {
		FixOutputBuffer(datalen + strlen(info) + 30);
		format_message(output_buffer, "300 DATA %s\n%s\n.\n", info, data);
	} else {
		FixOutputBuffer(datalen + 30);
		format_message(output_buffer, "300 DATA s %d\n%s\n.\n", datalen, data);
	}

	WriteLog(data);
	return CachedOutput(output_buffer);
}


int DebugBlock(const char *message) {
	if (!bf_debug || !message || !did_hello)
		return 1;
	FixOutputBuffer(strlen(message) + 15);
	sprintf(output_buffer, "302 DEBUG\n%s\n.\n", message);
	return io_puts(output_buffer);
}

int BFCreateDirectory(char *root, char *dir, int mode)
{
	char	*buf;

	CleanPath(root);
	errno = 0;
	if (platform_chdir(root) != 0) {
		buf = calloc(1, BFBUFSIZ);
		format_message(buf, bfCat(E_MKDIR_CDROOT), root, errno);
		DataBlock(buf, NULL);
		free(buf);
		return -1;
	}
	return BFCreateDirectoryRelative(dir, mode);
}

int BFCreateDirectoryRelative(char *dir, int mode)
{
	char    *buf, *p, *usedir, *root;
	char    *bf_rptr=NULL;
	char    *bf_sptr=NULL;
	wchar_t *tmpbuf;

	if (mode == 0)
		mode = 0755;
	root = calloc(sizeof(*root), BFBUFSIZ);
	buf = calloc(sizeof(*buf), BFBUFSIZ);
	tmpbuf = calloc(sizeof(*tmpbuf), BFBUFSIZ);

	platform_getcwd(root, BFBUFSIZ);
	CleanUnixPath(root);
	strcpy(buf, root);
	usedir = strdup(dir);
	CleanUnixPath(usedir);
	p = bf_strtok(usedir, "/", &bf_rptr, &bf_sptr);
	while (p) {
		strcat(buf, "/");
		strcat(buf, p);
		if (!isDir(buf) && platform_mkdir(buf, mode) != 0)
			break;
		p = bf_strtok(NULL, "/", &bf_rptr, &bf_sptr);
	}
	bf_strtok(NULL, NULL, NULL, &bf_sptr);
	free(usedir);
	free(root);
	free(tmpbuf);
	if (isDir(buf)) {
		free(buf);
		return 0;
	}
	free(buf);
	return -1;
}

/* 
 *  function to test a pathname to see if it's a directory. 
 *  doesn't alter path at all. 
 */
#ifndef S_ISDIR
#define S_ISDIR(x)	((x & S_IFDIR) == S_IFDIR)
#endif
int isDir(const char *path)
{
	struct 	stat sb;
	return (platform_stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)) ? 1 : 0;
}

/* 
 *  function to test a pathname to see if it's a regular file. 
 *  doesn't alter path at all. 
 */
#ifndef S_ISREG
#define S_ISREG(x)	((x & S_IFREG) == S_IFREG)
#endif
int isFile(const char *path)
{
	struct 	stat sb;
	return (platform_stat(path, &sb) == 0 && S_ISREG(sb.st_mode)) ? 1 : 0;
}

static const char *hex_code = "0123456789ABCDEF0123456789abcdef";
char *EncodeBlock(const unsigned char *src, int bytes, int len) {
	const unsigned char	*end = &src[bytes];
	unsigned char		*result, *p;
	size_t				size;
	int 				i;

	len >>= 1;
	size = (bytes << 1) + (bytes / len) + 2;
	result = p = malloc(size);

	for (i=0; src<end; i++) {
		if (i >= len) {
			i = 0;
			*p++ = '\n';
		}
		*p++ = hex_code[(*src >> 4) & 0x0F];
		*p++ = hex_code[*src++ & 0x0F];
	}
	*p = '\0';
	return (char *)result;
}

int DecodeBlock(unsigned char *result, const unsigned char *src) {
	int 		i;
	const char	*p;

	for (i=0; *src; i++) {
		if (*src == '\n') {
			do { src++; } while (*src == '\n');
			if (!*src) break;
		}

		p = strchr(hex_code, *src++);
		if (p == NULL) {
			result[i] = '?';
			if (*src++) continue;
			break;
		}
		result[i] = ((p - hex_code) << 4) & 0xF0;

		p = strchr(hex_code, *src++);
		if (p == NULL) {
			result[i] = '?';
			continue;
		}
		result[i] |= (p - hex_code) & 0xF;
	}
	return i;
}


void sayHello(int mode) {
	const char *msg;
	char buf[BFBUFSIZ >> 4];

	switch (mode) {
		case CODE_HELLO:            msg = CODE_HELLO_MSG;            break;
		case CODE_SECURE_HELLO:     msg = CODE_SECURE_HELLO_MSG;     break;
		case CODE_CERTIFIED_HELLO:  msg = CODE_CERTIFIED_HELLO_MSG;  break;
		case CODE_GOODBYE:          msg = CODE_GOODBYE_MSG;          break;
		default: msg = "???";
	}
	if (mode != CODE_GOODBYE) {
		did_hello = 1;
	}
	format_message(buf, "%d %s\n", mode, msg);
	io_puts(buf);
}


#ifdef WIN32
#define	FTW_F	1
#define	FTW_D	2
#define	FTW_DNR	3
#define	FTW_SL	4
#define	FTW_NS	5
#else
#if HAVE_FTW_H
#include <ftw.h>
#undef HAVE_FTS_H
#endif

#ifdef HAVE_FTS_H
#include <fts.h>
#define FTW_F		FTS_F
#define FTW_D		FTS_D
#define FTW_SL		FTS_SL
#define FTW_NS		FTS_NS
#endif

#endif


/* function to transmit a file/entity */
int SendFunction(const char *p, const struct stat *sb, int flags)
{
	int 	err, fd;
	size_t	bytes;
	off_t   offset;
	char	*buf, *info, *encoded;
	char	*path = CleanPath(strdup(p));

	err = 0;
	info = calloc(1, BFBUFSIZ);
	switch (flags) {
		case FTW_F:
			buf = calloc(1, BFXFERBLOCKSIZE);
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd != -1) {
				offset = (off_t)0;
				while ((bytes = read(fd, buf, BFXFERBLOCKSIZE)) > 0) {
					sprintf(info, "FILE %llu %ld %o %ld %ld %s",
							(unsigned long long)offset, (long)bytes, (int)sb->st_mode,
							(long)sb->st_size, (long)sb->st_mtime, path);
					encoded = EncodeBlock((unsigned char *)buf, bytes, XFERLINELEN);
					DataBlock(encoded, info);
					free(encoded);
					offset += bytes;
					memset(buf, 0, BFXFERBLOCKSIZE);
				}
				close(fd);
			}
			free(buf);
			break;
		case FTW_D:
#ifdef FTW_DNR
		case FTW_DNR:
#endif
			sprintf(info, "DIR %o %ld %s", (int)sb->st_mode, (long)sb->st_mtime, path);
			DataBlock(NULL, info);
			break;
		case FTW_SL:
			break;
		default:
			break;
	}
	free(info);
	free(path);
	return 0;
}

/* function to transmit a checkpoint for a file/entity */
int CheckpointFunction(const char *p, const struct stat *sb, int flags)
{
	char	*info, *buf;
	uint32_t crc;
	md5_state_t	ctx;
	md5_byte_t	digest[16];
	int 	fd;
	size_t	bytes;
	char	*path = CleanPath(strdup(p));

	info = calloc(1, BFBUFSIZ);
	crc = get_crc_string(path);
	switch (flags) {
		case FTW_F:
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd != -1) {
				md5_init(&ctx);
				buf = calloc(1, BFBUFSIZ);
				while ((bytes = read(fd, buf, BFBUFSIZ-1)) > 0)
					md5_append(&ctx, (const md5_byte_t *)buf, bytes);
				close(fd);
				md5_finish(&ctx, digest);
				for (fd = 0; fd < 16; fd++)
					sprintf(buf + (fd << 1), "%02x", digest[fd]);
				sprintf(info, "FILE %lx %ld %s", (unsigned long)crc, (long)sb->st_mtime, path);
				DataBlock(buf, info);
				free(buf);
			}
			else {
				free(path);
				return 0;
			}
			break;
		case FTW_D:
#ifdef FTW_DNR
		case FTW_DNR:
#endif
			sprintf(info, "DIR %lx %ld %s", (unsigned long)crc, (long)sb->st_mtime, path);
			DataBlock(NULL, info);
			break;
		case FTW_SL:
			sprintf(info, "SYMLINK %lx %ld %s", (unsigned long)crc, (long)sb->st_mtime, path);
			DataBlock(NULL, info);
			break;
		default:
			break;
	}
	free(info);
	free(path);
	return 0;
}

/* CRC-32-MPEG2 (aka IEEE 802.3 / V.42) */
static uint32_t *crc_table = NULL;
void gen_crc_table(void)			/* build the crc table */
{
	uint32_t crc, poly = 0xEDB88320L;
	int	i, j;

	if (NULL != crc_table)
		return;
	crc_table = calloc(256, sizeof(uint32_t));

	for (i = 0; i < 256; i++) {
		crc = i;
		for (j = 8; j > 0; j--) {
			if (crc & 1)
				crc = (crc >> 1) ^ poly;
			else
				crc >>= 1;
		}
		crc_table[i] = crc;
	}
}

uint32_t get_crc_string(char *s)	/* calculate the crc value */
{
	register uint32_t crc = 0xFFFFFFFF;
	if (s) {
		while (*s)
			crc = (crc >> 8) ^ crc_table[(crc ^ *s++) & 0xFF];
	}
	return crc ^ 0xFFFFFFFF;
}

uint32_t get_crc_file(char *path)	/* calculate the crc value */
{
	register uint32_t crc = 0xFFFFFFFF;
	register char *p;
	int fd;
	size_t	bytes;
	char	*buf;

	fd = platform_open2(path, O_RDONLY|O_BINARY);
	if (fd == -1) return 0;

	buf = malloc(BFBUFSIZ);
	while ((bytes = read(fd, buf, BFBUFSIZ-1)) > 0) {
		buf[bytes] = 0;
		for (p = buf; *p; p++)
			crc = (crc>>8) ^ crc_table[(crc ^ *p++) & 0xFF ];
	}
	close(fd);
	free(buf);
	return crc ^ 0xFFFFFFFF;
}

void cleanMyCruft(char *path) {
	if (path) {
		if (BFCONF_EXISTS(leave_tmp_file)) {
			char msg[BFBUFSIZ];
			sprintf(msg, bfCat(M_TMP_FILE_IS), path);
			DataBlock(msg, NULL);
		} else {
			platform_unlink(path);
		}
	}
}

void StoreUserData(char *data, size_t bytes) {
	char	*e;
	FILE	*f;

	if (data && NULL != (e = platform_getenv("_USERDATA"))) {
		if (NULL != (f = platform_fopen(e, "w"))) {
			fwrite(data, 1, bytes, f);
			fclose(f);
		}
		platform_freeenv(e);
	}
}

void FreeUserData(void) {
	char	*e;

	if (NULL != (e = platform_getenv("_USERDATA"))) {
		platform_unlink(e);
		platform_freeenv(e);
	}
}



/* The magic_login is intended to increase the security of the agent.
 * Its behavior depends on the agent configuration, as follows:
 *
 * Running as root (Unix) or a member of the Administrators group (Windows):
 * - magic_login is ignored
 * - user must provide valid login credentials
 *
 * Running as an ordinary user:
 * - The credentials must match the magic login
 * - If no magic login is defined, then the credentials must be left blank
 *
 * Return value:
 * 		SU_OK           = Success
 * 		SU_OK_NOACCOUNT = Not applicable (no magic_login defined)
 * 		SU_ILLEGAL      = Denied - no login specified
 * 		SU_UNKNOWNUSER  = Denied - username or password mismatch
 */
int auth_magic = 0;
int check_magic_login(const char *username, const char *password) {
	char *magic_login, *p;
	int  ret = SU_UNKNOWNUSER;

	magic_login = bfconf_get("magic_login", NULL);
	if (magic_login == NULL) return SU_OK_NOACCOUNT;

	if (username == NULL || *username == '\0' ||
			password == NULL || *password == '\0')
	{
		free(magic_login);
		return SU_ILLEGAL;
	}

	p = strrchr(magic_login, ':');
	if (p != NULL) *p++ = '\0';

	if (strcmp(username, magic_login) == 0) {
		size_t len = strlen(password);
#ifdef ZOS
		char *pass = strdup(password);
		__etoa_l(pass, len);
		password = pass;
#endif
		if (smd5_check(p, (const md5_byte_t *)password, len) == 0) {
			auth_magic = 1;
			ret = SU_OK;
		}
#ifdef ZOS
		free(pass);
#endif
	}

	free(magic_login);
	return ret;
}

