/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include <stdarg.h>

#undef DEBUG_FORMAT_MESSAGE
#ifdef DEBUG_FORMAT_MESSAGE
#  define DFM(x) x
#else
#  define DFM(x) /**/
#endif
typedef enum {
	FT_UNUSED = 0,
	FT_C,
	FT_S,
	FT_P,
	FT_D,
	FT_LD,
	FT_O,
	FT_U,
	FT_LU,
	FT_X,
	FT_X_UPPER
} format_type_t;

typedef union {
	char c;
	const char *s;
	const void *p;
	int d;
	long ld;
	unsigned int u;
	unsigned long lu;
} format_data_t;

#define MAX_MSG_ARGS 10
#define MAX_MSG_FIELDS 24
/* This is a bit like sprintf, but with limited functionality.
 * Only the c, s, p, d, ld, u, and lu formats are supported.
 * We call sprintf to do the real work for all formats except
 * %s.  The positional parameter convention of %1$s, %2$ld, etc.
 * is supported, but a maximum of 9 arguments may be used in any
 * case.  Flags, width, precision, '*', and formats other than
 * those listed above will produce the message "[!!!c]", where
 * 'c' is the character that was not understood.
 *
 * The obvious question is "Why we are doing this?"  The answer is
 * that sprintf is generally aware of your locale.  This will not
 * be a problem in most circumstances, but (on Windows in particular)
 * sprintf might do the wrong thing when passed a UTF-8 format
 * string.  The last byte of the UTF-8 literal may be a DBCS lead
 * byte and swallow up the '%' that follows it, leading to the use
 * of the wrong arguments and potentially a crash.
 *
 * There does not appear to be any reliable and portable way to
 * avoid this problem except by using something other than sprintf
 * to handle the literal UTF-8 data.  Fortunately, since we control
 * the format strings, we don't have to completely reimplement
 * sprintf, here -- we only implement the functionality that we
 * actually need/use.  If at any point we see something we don't
 * understand, we pass the whole thing off to vsprintf just in
 * case.
 */
int format_message(char *buf, const char *msg, ...) {
	const char *lit_beg[MAX_MSG_FIELDS];
	int lit_len[MAX_MSG_FIELDS];
	int field_arg[MAX_MSG_FIELDS];
	format_data_t arg_data[MAX_MSG_ARGS];
	format_type_t arg_type[MAX_MSG_ARGS];
	const char *m = msg, *mark;
	char *b = buf;
	int maxarg = 0, fields = 0;
	int i, arg = 0, len;
	va_list va;

	memset(arg_type, 0, sizeof(arg_type));
	va_start(va, msg);

	/* Step 1: Parse message format and count args */
	while (NULL != (m = strchr(mark = m, '%'))) {
		/* Step 1a: Save range of literal text before % */
		DFM(printf("format_message: field=%d lit_len=%d lit_beg->%p[%s]\n",
					fields, (int)(m-mark), mark, mark));
		lit_beg[fields] = mark;
		lit_len[fields] = m - mark;

		/* Step 1b: Deal with %% by using the first % in the literal */
		/*   and not consuming an arg. */
		if (*++m == '%') {
			DFM(printf("format_message: %%%%\n"));
			++lit_len[fields];
			++m;
			field_arg[fields++] = 0;
			continue;
		}

		/* Step 1c: Figure out which arg this is */
		if (*m > '0' && *m <= '9') {
			arg = *m++ - '0';
			if (*m != '$') {
				DFM(printf("format_message: 1c: Numbers, but no $: [%s]\n", msg));
				return vsprintf(buf, msg, va);
			}
			DFM(printf("format_message: positional: arg=%d\n", arg));
			++m;
		} else {
			++arg;
			DFM(printf("format_message: ordered: arg=%d\n", arg));
		}
		if (maxarg < arg)
			maxarg = arg;
		field_arg[fields++] = arg;

		/* Step 1d: Decode the type.  Only c, s, d, u, ld, and lu are allowed. */
		switch (*m++) {
			case 's':  arg_type[arg] = FT_S;  break;
			case 'd':
			case 'i':  arg_type[arg] = FT_D;  break;
			case 'l':
				switch (*m++) {
					case 'd':  arg_type[arg] = FT_LD;  break;
					case 'u':  arg_type[arg] = FT_LU;  break;
					default: return vsprintf(buf, msg, va);
				}
				break;
			case 'c':  arg_type[arg] = FT_C;  break;
			case 'u':  arg_type[arg] = FT_U;  break;
			case 'o':  arg_type[arg] = FT_O;  break;
			case 'x':  arg_type[arg] = FT_X;  break;
			case 'X':  arg_type[arg] = FT_X_UPPER;  break;
			case 'p':  arg_type[arg] = FT_P;  break;
			default:
				DFM(printf("format_message: 1d: bad arg_type during parse: [%s]\n", msg));
				return vsprintf(buf, msg, va);
		}
		DFM(printf("format_message: arg type='%c' -> %d\n", m[-1], arg_type[arg]));
	}

	/* Step 2: Take care of the leftover literal.  If that's all we have, then
	 *   all we need to do is copy the message and be done with it.
	 */
	if (0 == fields) {
		DFM(printf("format_message: literal only\n"));
		len = strlen(msg);
		memcpy(buf, msg, len + 1);
		return len;
	}
	DFM(printf("format_message: literal assigned to field=%d\n", fields));
	lit_beg[fields] = mark;
	lit_len[fields] = strlen(mark);
	field_arg[fields++] = 0;

	/* Step 3: Read in the args */
	for (arg=1; arg<=maxarg; ++arg) {
		DFM(printf("format_message: reading arg=%d type=%d\n", arg, arg_type[arg]));
		switch (arg_type[arg]) {
			case FT_S:
				arg_data[arg].s = va_arg(va, const char *);
				break;
			case FT_D:
			case FT_C:
				arg_data[arg].d = va_arg(va, int);
				break;
			case FT_LD:
				arg_data[arg].ld = va_arg(va, long);
				break;
			case FT_U:
			case FT_O:
			case FT_X:
			case FT_X_UPPER:
				arg_data[arg].u = va_arg(va, unsigned int);
				break;
			case FT_LU:
				arg_data[arg].lu = va_arg(va, unsigned long);
				break;
			case FT_P:
				arg_data[arg].p  = va_arg(va, const void *);
				break;
			default:
				DFM(printf("format_message: 3: Bad arg_type during output: [%s]\n", msg));
				return vsprintf(buf, msg, va);
		}
	}
	va_end(va);

	/* Step 4: Output phase */
	for (i=0; i<fields; ++i) {
		int arg = field_arg[i];
		DFM(printf("format_message: writing field #%d lit_beg=%d lit_len=%d\n",
					i, lit_beg[i] - msg, lit_len[i]));
		len = lit_len[i];
		if (len > 0) {
			memcpy(b, lit_beg[i], len);
			b += len;
		}
		if (arg > 0) {
			switch (arg_type[arg]) {
				case FT_S:   strcpy(b, arg_data[arg].s); b += strlen(b);  break;
				case FT_D:   b += sprintf(b, "%d",  arg_data[arg].d );    break;
				case FT_C:   *b++ = (char)arg_data[arg].d;                break;
				case FT_LD:  b += sprintf(b, "%ld", arg_data[arg].ld);    break;
				case FT_U:   b += sprintf(b, "%u",  arg_data[arg].u );    break;
				case FT_O:   b += sprintf(b, "%o",  arg_data[arg].u );    break;
				case FT_X:   b += sprintf(b, "%x",  arg_data[arg].u );    break;
				case FT_X_UPPER:
							 b += sprintf(b, "%X",  arg_data[arg].u );    break;
				case FT_LU:  b += sprintf(b, "%lu", arg_data[arg].lu);    break;
				case FT_P:   b += sprintf(b, "%p",  arg_data[arg].p );    break;
				default:
					/* PARANOID */
					DFM(printf("format_message: 4: Bad format: [%s]\n", msg));
					return vsprintf(buf, msg, va);
			}
		}
	}
	*b = '\0';
	DFM(printf("format_message: ret=%d\n", b - buf));
	return b - buf;
}


/* memrchr is usually only available on GNU systems. */
#ifndef HAVE_MEMRCHR
void *bf_memrchr(const void *s, int c, size_t n)
{
	const uint8_t *start = s;
	const uint8_t *p = &start[n];
	while (--p >= start) {
		if (*p == (uint8_t)c)
			return (void *)p;
	}
	return NULL;
}
#endif

/*
 * A (slow) implementation of strdup(3), just in case your C library
 * does not have it.
 */
#ifndef HAVE_STRDUP
char *BFStrDup(const char *src)
{
	size_t len = strlen(src) + 1;
	char *dst = malloc(len);
	memcpy(dst, src, len);
	return dst;
}
#endif /* HAVE_STRDUP */

#ifndef LANG_TEST

/*
 *
 * function to compare env strings in a platform dependant way.
 *
 */
int BFStrCmp(const char *a, const char *b, int len) {
	char	*src, *dst;
	int		result;

	src = calloc(1, len+1);
	strncpy(src, a, len);
	dst = calloc(1, len+1);
	strncpy(dst, b, len);

	if (platform_style() == STYLE_WINDOWS) {
		LowerString(src);
		LowerString(dst);
	}
	result = strcmp(src, dst);
	free(src);
	free(dst);
	return result;
}

/*
 *  function to lowercase a string.
 *  modifies the original string and returns it.
 */
char *LowerString(char *src)
{
	if (src) {
		unsigned char *p = (unsigned char *)src;
		while (*p) {
			*p = (unsigned char)tolower(*p);
			++p;
		}
	}
	return src;
}

/*
 *  function to switch all \'s to /'s
 *  modifies the original string and returns it.
 */
char *toForwardSlashes(char *src)
{
	if (src) {
		char *p = src;
		while (NULL != (p = strchr(p, '\\')))
			*p++ = '/';
	}
	return src;
}

/*
 *  function to switch all /'s to \'s
 *  modifies the original string and returns it.
 */
char *toBackSlashes(char *src)
{
	if (src) {
		char *p = src;
		while (NULL != (p = strchr(p, '/')))
			*p++ = '\\';
	}
	return src;
}

/*
 *  function to trim ending newline chars
 *  modifies the original string and returns it.
 *  won't chop a string with an embedded newline
 *  that doesn't end in a newline.
 */
char *RemoveTrailingNewline(char *src)
{
	if (src && *src) {
		char *p = &src[strlen(src) - 1];
		while (p >= src && (*p == '\r' || *p == '\n'))
			*p-- = '\0';
	}
	return src;
}

/*
 *  function to trim trailing newline chars
 *  modifies the original string and returns it.
 *  will chop a string on the first newline.
 */
char *RemoveNewline(char *src)
{
	char	*p;

	if (src && *src) {
		if ((p = strrchr(src, '\r')))
			*p = 0;
		if ((p = strrchr(src, '\n')))
			*p = 0;
	}
	return src;
}

/*
 *  function to see if there's an EOL sequence in a string
 *  since any platform will have a NL as part of EOL, we'll
 *  just look for that char. this could be enhanced to handle
 *  EOL differently on windows/unix.
 */
char *HasEOL(char *src)
{
	return strchr(src, '\n');
}

/*
 *  function to change repeated /'s into a single /
 *  modifies the original string and returns it.
 */
char	*CollapseSlashes(char *src)
{
	char	*original_string = src, *p;

	if (!src || !*src)
		return src;

	while (*src) {
		if (*src == '/' && ++src) {
			p = src;
			while (*p == '/')
				++p;
			if (p != src)
				strcpy(src, p);
		}
		else
			src++;
	}

	return original_string;
}

/* strip leading \'s off of a pathname when there is a drive letter */
/* involved. If there doesn't appear to be a drive letter, then make */
/* sure there are no more than two \'s leading the way */
/* we have to handle windows UNC pathnames as well, which are tedious */
/*
 * Valid:     e:\foo\bar
 * Valid:     \\foo\bar
 * Invalid:   \e:\foo\bar
 * Invalid:   \\foo:\bar
 * Invalid:   \\foo (unc must have 2 components)
 *
 * also clean up any double slashes left over in windows path names
 */
char	*clean_win_slashes(char *src) {
	char	*p;

	if (!src || !*src)
		return src;

	if ((p = strchr(src, ':')) != NULL) {
		if (p == src)
			return src;	/* a path that starts with a ':' ??? */

		/* strip out doubled post drive \'s, i.e.:   e:\\dir */
		/* p points to the :. if that's it, then strncmp will fail and we'll drop out */
		while (strncmp(p+1, "\\\\", 2) == 0)
			strcpy(p+1, p+2);

		p--;
		if (toupper(*p) < 'A' || toupper(*p) > 'Z')
			return src;	/* an illegal drive letter character */

		strcpy(src, p);
	}
	else {
		while (src[0] == '\\' && src[1] == '\\' && src[2] == '\\')
			strcpy(src, src+1);
		if (strncmp(src, "\\\\", 2) == 0 && strchr(&src[2], '\\') == NULL)
			strcpy(src, src+1);
	}

	/* Clean up any remaining double slashes. after the first character. */
	p = &src[1];
	while (*p) {
		if (strncmp(p, "\\\\", 2) == 0)
		    	strcpy(p, p+1);
		else
			++p;
	}
	return src;
}

/*
 *  function to chomp trailing /'s, as well as \r\n's
 *  modifies the original string and returns it.
 */
char	*ChompSlash(char *src)
{
	int	i;
	char	sep;

	if (!src || !*src)
		return src;

	sep = PathSeparator();
	src = RemoveNewline(src);
	for (i = strlen(src) - 1 ; src && i && src[i] == sep ; i--)
		src[i] = 0;
	/*  if src == "/", do we make it an empty string?? */
	return src;
}

/*
 *  function to clean up a string and make it a
 *  nice unix pathname.
 *  modifies the original string and returns it.
 */
char	*CleanUnixPath(char *src)
{
	return (!src || !*src)
		? src
		: ChompSlash(CollapseSlashes(toForwardSlashes(src)));
}

/* strip leading \'s off of a pathname when there is a drive letter */
/* involved. If there doesn't appear to be a drive letter, then make */
/* sure there are no more than two \'s leading the way */
/* we have to handle windows UNC pathnames as well, which are tedious */
/*
 * Valid:     e:\foo\bar
 * Valid:     \\foo\bar
 * Invalid:   \e:\foo\bar
 * Invalid:   \\foo:\bar
 */
char	*clean_win_leading_slashes(char *src) {
	char	*p, *q;

	if (!src || !*src)
		return src;

	if ((p = strchr(src, ':')) != NULL) {
		if (p == src)
			return src;	/* a path that starts with a ':' ??? */

		/* strip out doubled post drive \'s, i.e.:   e:\\dir */
		/* p points to the :. if that's it, then strncmp will fail and we'll drop out */
		while (strncmp(p+1, "\\\\", 2) == 0)
			strcpy(p+1, p+2);

		p--;
		if (toupper(*p) < 'A' || toupper(*p) > 'Z')
			return src;	/* an illegal drive letter character */

		if (p <= src)		/* protect from backing up too far later on */
			return src;

		/* back over the lead of the string, it should be nothing but \'s */
		q = p--; /* leave q pointing to the drive letter */
		while (p > src && *p == '\\')
			--p;

		if (p == src) /* we have success */
			strcpy(src, q);
		return src;
	}
	else
		while (src[0] == '\\' && src[1] == '\\' && src[2] == '\\')
			strcpy(src, src+1);

	p = &src[1];
	while (*p) {
		if (*p == '\\' && *(p+1) == '\\')
			strcpy(p, p+1);
		else
			++p;
	}
	return src;
}

/*
 *  function to clean up a string and make it a
 *  nice windows pathname.
 *  modifies the original string and returns it.
 */
char	*CleanWindowsPath(char *src)
{
	return (!src || !*src)
		? src
		: clean_win_slashes(ChompSlash(toBackSlashes(src)));
}

/*
 * function to clean up a path for the current platform
 *
 */
char	*CleanPath(char *src)
{
	return (!src || !*src)
		? src
		: ((platform_style() == STYLE_WINDOWS
			? CleanWindowsPath(src)
			: CleanUnixPath(src)));
}

char PathSeparator(void)
{
	switch (platform_style()) {
		case STYLE_WINDOWS:
			return '\\';
		case STYLE_VOS:
			return '?';
		case STYLE_UNIX:
		default:
			return '/';
	}
}

char EnvPathSeparator(void)
{
	switch (platform_style()) {
		case STYLE_WINDOWS:
			return ';';
		default:
			return ':';
	}
}

char *shiftArgs(char **p) {
	int	i;
	char	*arg;

	arg = p[0];
	for (i = 0 ; p[i] ; i++)
		p[i] = p[i+1];
	return arg;
}

void freeArgs(char **p) {
	int	i;

	if (!p)
		return;
	for (i = 0 ; p[i] ; i++)
		free(p[i]);
	free(p);
}

char *nextQuote(char *src, char ch) {
	while (*src) {
		if (*src == ch)
			return src;
		if (*src == '\\')
			strcpy(src, src+1);
		src++;
	}
	return NULL;
}


char **getArgs(char *cl) {
	char	**result, *p;
	int	i;

	if (!cl || *cl == '\0')
		return NULL;

	result = calloc(sizeof(char *), 128);

	i = 0;
	while (*cl && i < 127) {
		/* skip over leading whitespace */
		while (*cl == ' ' || *cl == '\t')
			++cl;

		/* find the end of the current arg */
		if (*cl == '"' || *cl == '\'') {
			if ((p = nextQuote(cl+1, *cl)) == NULL) {
				result[i++] = strdup(cl+1);
				return result;
			}
			++cl;
		}
		else {
			if ((p = strpbrk(cl, " \t\n")) == NULL) {
				result[i++] = strdup(cl);
				return result;
			}
		}
		*p++ = 0;
		result[i++] = strdup(cl);
		cl = p;
	}
	return result;
}


char **getArgsN(char *cl, int num) {
    char    **result, *p;
    int     i;

    if (!cl || *cl == '\0')
        return NULL;

    result = calloc(sizeof(char *), 128);

    i = 0;
    while (*cl && i < 127) {
        /* skip over leading whitespace */
        while (*cl == ' ' || *cl == '\t')
            ++cl;

        /* find the end of the current arg */
        if (*cl == '"' || *cl == '\'') {
            if ((p = nextQuote(cl+1, *cl)) == NULL) {
                result[i++] = strdup(cl+1);
                return result;
            }
            ++cl;
        }
        else {
            if ((p = strpbrk(cl, " \t\n")) == NULL) {
                result[i++] = strdup(cl);
                return result;
            }
        }
        *p++ = 0;
        result[i++] = strdup(cl);
        cl = p;
        if (num>=0) {num--;}
        if (num == 0) {result[i++]=strdup(cl);break;}
    }
    return result;
}


char **getArgsC(char *cl, char c) {
    char    **result, *p;
    int     i;

    if (!cl || *cl == '\0')
        return NULL;

    result = calloc(sizeof(char *), 128);

    i = 0;
    while (*cl && i < 127) {
        /* skip over leading whitespace */
        while (*cl == ' ' || *cl == '\t')
            ++cl;

        /* find the end of the current arg */
        if (*cl == '"' || *cl == '\'') {
            if ((p = nextQuote(cl+1, *cl)) == NULL) {
                result[i++] = strdup(cl+1);
                return result;
            }
            ++cl;
        }
        else {
            if ((p = strchr(cl, c)) == NULL) {
                result[i++] = strdup(cl);
                return result;
            }
        }
        *p++ = 0;
        result[i++] = strdup(cl);
        cl = p;
    }
    return result;
}

/* part of some string based on content */
char *bf_strpart(char *pstr, char *psub) {
	char *temp;
	char *newstr;

	if (!pstr)
		return NULL;

	if (psub && *psub != '\0') {
		if ((temp = strstr(pstr, psub))) {
			newstr = calloc(1, temp - pstr + 1);
			strncpy(newstr, pstr, temp - pstr);
			return newstr;
		}
	}
	return strdup(pstr);
}

regex_t *bf_regex_compile(char *pattern, int icase) {
	regex_t	*buf = calloc(1, BFBUFSIZ);
	if (regcomp(buf, pattern, (icase ? REG_ICASE : 0) | REG_EXTENDED) != 0)
		return NULL;
	return buf;
}

void bf_regex_free(regex_t *srch) {
	regfree(srch);
}

int bf_regex_prep(
	regex_t **lhs,
	char **spat,
	char **rhs,
	char **target,
	char *buf)
{
	int		icase;
	char	*repl, *srch, *src, delim;

	/* first char is the delimiter character */
	*lhs = NULL;
	*rhs = *target = NULL;
	srch = buf;
	delim = *srch++;

	/* srch is the start of the regex, now we need to find the start of the replacement */
	/* scan the search pattern for the ending delimiter */
	for (repl = srch ; *repl && *repl != delim ; repl++) {
		if (*repl == '\\') {
			++repl;
			if (!*repl)
				break;
		}
	}
	if (!*repl || (repl == srch)) {
		*target = strdup(bfCat(E_EDIT_TERM_SEARCH));
		return -1;
	}

	/* we found the start of the replacement pattern */
	*repl++ = 0;
	*spat = strdup(buf);

	/* scan the replacement pattern for the ending delimiter */
	for (src = repl ; *src && *src != delim ; src++) {
		if (*src == '\\') {
			++src;
			if (!*src)
				break;
		}
	}
	if (!*src) {
		*target = strdup(bfCat(E_EDIT_TERM_REPLACE));
		free(*spat);
		*spat = NULL;
		return -1;
	}
	*src++ = 0;
	icase = 0;

	/* check for the ignore case flag */
	if (tolower(*src) == 'i')
		icase = 1;
	icase = 0; /* icase flag disabled for now */

	/* skip over any flags */
	while (*src && !isspace(*src & 0xFF))
		src++;

	/* eat up source leading whitespace */
	while (*src && isspace(*src & 0xFF))
		src++;

	/* now expand the source and replacement string for env vars */
	*rhs = env_expand_var(repl, 1, 0);
	*target = env_expand_var(src, 0, 0);

	/* now compile the search string */
	srch = env_expand_var(srch, 0, 0);
	*lhs = bf_regex_compile(srch, icase);
	free(srch);
	if (NULL == *lhs) {
		free(*spat);
		free(*rhs);
		free(*target);
		*spat = *rhs = NULL;
		*target = strdup(bfCat(E_EDIT_BAD_EXPRESSION));
		return -1;
	}
	return 0;
}

char *bf_regex_apply(const char *buf, const regex_t *srch, const char *repl) {
	regmatch_t  matches[MAX_REGEX_SUBSTRINGS];
	int         i, j, width;
	const char  *p;
	char        *out, *q, *result, num[2];

	/* srch points to a null terminated compiled search pattern, repl points to a null */
	/* terminated expanded replacement pattern, src points to the expanded source string */

	if ((i = regexec(srch, buf, MAX_REGEX_SUBSTRINGS, matches, 0)) != 0)
		return strdup(buf);

	/* parse the replacement string to substitute \# replacements */
	num[1] = 0;
	out = q = calloc(1, BFBUFSIZ);
	for (p = repl; *p; ++p) {
		if (*p != '\\') {
			*q++ = *p;
			continue;
		}
	
		++p;
		if (*p == '\0')
			break;

		/* Check for and handle backrefs like \1 */
		num[0] = *p;
		j = atoi(num);
		if (j >= 1 && j <= 9) {
			if (matches[j].rm_so > -1 && matches[j].rm_eo > -1) {
				width = matches[j].rm_eo - matches[j].rm_so;
				strncat(out, &buf[matches[j].rm_so], width);
				q += width;
			}
			continue;
		}

		/* wasn't a \1 pattern, so just dequote it */
		switch (*p) {
			case 'a': *q++ = '\a'; break;
			case 'b': *q++ = '\b'; break;
			case 'f': *q++ = '\f'; break;
			case 'n': *q++ = '\n'; break;
			case 'r': *q++ = '\r'; break;
			case 't': *q++ = '\t'; break;
			default: *q++ = *p;
		}
	}

	/* out now has the parsed replacement string! */
	/* go back and substitute the whole replacement string into the original string */
	result = calloc(1, strlen(buf) + strlen(out) + 1);
	strcpy(result, buf);
	result[matches[0].rm_so] = 0;
	strcat(result, out);
	strcat(result, &buf[matches[0].rm_eo]);
	free(out);
	return result;
}

/*	OpenBSD: strlcpy.c,v 1.8 2003/06/17 21:56:24 millert Exp	*/

/*
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
#ifndef HAVE_STRLCPY
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return s - src - 1;	/* count does not include NUL */
}
#endif


/*
 * bf_strtok allows the use of strtok_r accross all
 * platforms with a cleaner interface.
 *
 * Called 3 ways:
 *
 *     - initial call (string, delim, &NULL, &NULL); sets up
 *       the dereferenced pointer and makes the initial
 *       call to strtok_r (3rd and 4th param: allocated
 *       char *bf_ptr=NULL; passed &bf_ptr)
 *
 *     - item call    (NULL, delim, &bf_rptr, &bf_sptr); the
 *       non-NULL values in the 3rd and 4th parameters cause
 *       the reuse of the dereferenced 3rd pointer and the
 *       traversing of the original string
 *
 *     - cleanup call (NULL, NULL, NULL, &bf_sptr); the three
 *       NULL parameter values indicate the cleanup mode.
 *
 * NOTE: If delim is NULL but either the 1st or 3rd argument is
 *   not NULL, then nothing happens, and NULL is returned.
 *
 * NOTE: The cleanup of the allocated memory happens automatically
 *   when the returned string is NULL (this means that an explicit
 *   cleanup call is not needed if the entire string is traversed)
 */

#ifndef HAVE_STRTOK_R
#include "strtok_r.c"
#endif

char *bf_strtok(char *s, const char *delim, char **ptrptr, char **saveptr) {
	char *lstr = NULL;

	if (delim == NULL) {
		if (s == NULL && ptrptr == NULL && *saveptr != NULL) {
			free(*saveptr);
			*saveptr = NULL;
		}
		return NULL;
	}

	if (*saveptr == NULL) {
		*saveptr = *ptrptr = calloc(1, strlen(s)+1);
		lstr = strtok_r(s, delim, ptrptr);
	} else {
		lstr = strtok_r(NULL, delim, ptrptr);
	}

	if (lstr == NULL && *saveptr != NULL) {
		free(*saveptr);
		*saveptr=NULL;
	}

	return lstr;
}

char *bf_expose_special_chars(char *str) {
	char	*p;

	p = calloc(1, BFBUFSIZ);
	while (*str) {
		switch (*str) {
			case '\n':
				strcat(p, "<LF>");
				break;
			case '\r':
				strcat(p, "<CR>");
				break;
			case '\t':
				strcat(p, "<^I>");
				break;
			case '\b':
				strcat(p, "<^H>");
				break;
			default:
				if (*str < ' ')
					sprintf(&str[strlen(str)], "%02x", *str);
				else
					p[strlen(p)] = *str;
				break;
		}
		str++;
	}
	return p;
}

#endif /* !LANG_TEST */
